#!/usr/bin/env python3

# Copyright © Advanced Micro Devices, Inc., or its affiliates.
# SPDX-License-Identifier:  MIT

import sys

if sys.version_info[0] < 3:
    raise Exception("Must be using Python 3")

import os
import argparse
from pathlib import Path
from ctypes import *
import ctypes
from copy import deepcopy
from trace_view import view_trace
import sys
import glob
import numpy as np
import stitch
import gc
from collections import defaultdict
import service

ATT_VERSION = 5

class pcInfo(ctypes.Structure):
    _fields_ = [
        ("addr", c_uint64),
        ("marker_id", c_uint64)
    ]
    def to_v2_pc(self):
        if self.marker_id == 0:
            return self.addr
        return self.addr | (self.marker_id<<service.ID_OFFSET) | (1<<service.HEADER_OFFSET)

class TraceData(ctypes.Structure):
    _fields_ = [
        ("type", c_uint64, 8),
        ("hitcount", c_uint64, 56),
        ("latency", c_uint64),
        ("pc", pcInfo)
    ]

class TraceDataTranslated:
    def __init__(self, inst):
        self.type = inst.type
        self.num_waves = inst.hitcount
        self.cycles = inst.latency
        if self.type == stitch.PCINFO:
            self.cycles = inst.pc.to_v2_pc()

class Trace:
    def __init__(self, traceid, tracesize, instructions_array):
        self.instructions = [TraceDataTranslated(instructions_array[k]) for k in range(tracesize)]
        self.traceid = traceid

class PerfEvent(ctypes.Structure):
    _fields_ = [
        ("time", c_uint64),
        ("event0", c_uint16),
        ("event1", c_uint16),
        ("event2", c_uint16),
        ("event3", c_uint16),
        ("cu", c_uint8),
        ("bank", c_uint8),
    ]

    def toTuple(self):
        return (
            int(self.time),
            int(self.event0),
            int(self.event1),
            int(self.event2),
            int(self.event3),
            int(self.cu),
            int(self.bank),
        )


class CodeWrapped(ctypes.Structure):
    """ Matches CodeWrapped on the python side """
    _fields_ = [('line', ctypes.c_char_p),
                ('loc', ctypes.c_char_p),
                ('to_line', ctypes.c_int),
                ('value', ctypes.c_int),
                ('index', ctypes.c_int),
                ('line_num', ctypes.c_int),
                ('addr', ctypes.c_int64)]


class KvPair(ctypes.Structure):
    """ Matches pair<int, int> = (key, value) on the python side """
    _fields_ = [('key', ctypes.c_int),
               ('value', ctypes.c_int)]


class ReturnAssemblyInfo(ctypes.Structure):
    """ Matches ReturnAssemblyInfo on the python side """
    _fields_ = [('code', POINTER(CodeWrapped)),
                ('jumps', POINTER(KvPair)),
                ('code_len', ctypes.c_int),
                ('jumps_len', ctypes.c_int)]


class WaveState(ctypes.Structure):
    _fields_ = [
        ("type", c_int32),
        ("duration", c_int32)
    ]

class WaveInstruction(ctypes.Structure):
    _fields_ = [
        ("time", c_int64),
        ("duration", c_int64)
    ]

class Wave(ctypes.Structure):
    _fields_ = [
        ("simd", ctypes.c_uint8),
        ("wave_id", ctypes.c_uint8),
        ("trap_status", ctypes.c_uint8),
        ("reserved", ctypes.c_uint8),

        # total VMEM/FLAT/LDS/SMEM instructions issued
        # VMEM Pipeline: instrs and stalls
        ("num_vmem_instrs", ctypes.c_int),
        ("num_vmem_stalls", ctypes.c_int),
        # FLAT instrs and stalls
        ("num_flat_instrs", ctypes.c_int),
        ("num_flat_stalls", ctypes.c_int),
        # LDS instr and stalls
        ("num_lds_instrs", ctypes.c_int),
        ("num_lds_stalls", ctypes.c_int),
        # SCA instrs stalls
        ("num_salu_instrs", ctypes.c_int),
        ("num_smem_instrs", ctypes.c_int),
        ("num_salu_stalls", ctypes.c_int),
        ("num_smem_stalls", ctypes.c_int),
        # Branch
        ("num_branch_instrs", ctypes.c_int),
        ("num_branch_taken_instrs", ctypes.c_int),
        ("num_branch_stalls", ctypes.c_int),

        # total issued memory instructions
        ("num_mem_instrs", ctypes.c_int),
        # total valus insts and stalls
        ("num_valu_stalls", ctypes.c_int),
        ("num_valu_instrs", ctypes.c_size_t),
        # total issued instructions (compute + memory)
        ("num_issued_instrs", ctypes.c_size_t),
        # Begin and end cycle
        ("begin_time", ctypes.c_int64),
        ("end_time", ctypes.c_int64),
        ("traceid", ctypes.c_int64),

        ("timeline_size", ctypes.c_size_t),
        ("instructions_size", ctypes.c_size_t),
        ("timeline_array", POINTER(WaveState)),
        ("instructions_array", POINTER(WaveInstruction)),
    ]


class PythonWave:
    def __init__(self, sourcew):
        for property, value in Wave._fields_:
            try:
                setattr(self, deepcopy(property), deepcopy(getattr(sourcew, property)))
            except:
                pass

        self.timeline = [
            (int(sourcew.timeline_array[k].type), int(sourcew.timeline_array[k].duration))
            for k in range(self.timeline_size)
        ]
        self.timeline_array = None

        self.instructions = [
            (int(sourcew.instructions_array[k].time), int(sourcew.instructions_array[k].duration))
            for k in range(self.instructions_size)
        ]
        self.instructions_array = None


class ReturnInfo(ctypes.Structure):
    _fields_ = [
        ("flags", ctypes.c_uint64),
        ("binaryID", ctypes.c_uint64),
        ("num_traces", ctypes.c_uint64),
        ("tracesizes", POINTER(ctypes.c_uint64)),
        ("traceIDs", POINTER(ctypes.c_int64)),
        ("tracedata", POINTER(POINTER(TraceData))),

        ("num_events", ctypes.c_uint64),
        ("perfevents", POINTER(PerfEvent)),
        ("occupancy", POINTER(ctypes.c_uint64)),
        ("num_occupancy", ctypes.c_uint64),
        ("kernel_id_addr", POINTER(pcInfo)),
        ("num_kernel_ids", ctypes.c_uint64),

        ("wavedata", POINTER(Wave)),
        ("num_waves", ctypes.c_uint64),
    ]


rocprofv2_att_lib = os.getenv("ROCPROFV2_ATT_LIB_PATH")
if rocprofv2_att_lib is None:
    print(
        "ATT Lib path not set. Use export ROCPROFV2_ATT_LIB_PATH=/path/to/librocprofv2_att.so"
    )
    quit()
path_to_parser = os.path.abspath(rocprofv2_att_lib)
SO = CDLL(path_to_parser)

SO.AnalyseBinary.restype = ReturnInfo
SO.AnalyseBinary.argtypes = [ctypes.c_char_p]
SO.wrapped_parse_binary.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
SO.wrapped_parse_binary.restype = ReturnAssemblyInfo
SO.FreeBinary.argtypes = [ctypes.c_uint64]

def parse_binary(filename, kernel=None):
    if kernel is None or kernel == "":
        kernel = ctypes.c_char_p(0)
    else:
        with open(glob.glob(kernel)[0], "r") as file:
            kernel = file.readlines()
        kernel = kernel[0].split(": ")[1].split(".kd")[0]
        kernel = str(kernel).encode("utf-8")
    filename = os.path.abspath(str(filename))
    info = SO.wrapped_parse_binary(str(filename).encode("utf-8"), kernel)

    code = []
    for k in range(info.code_len):
        code_entry = info.code[k]

        line = deepcopy(code_entry.line.decode("utf-8"))
        loc = deepcopy(code_entry.loc.decode("utf-8"))

        to_line = int(code_entry.to_line) if (code_entry.to_line >= 0) else None
        loc = loc if len(loc) > 0 else None

        # asm, inst_type, addr, loc, index, line_num, hitcount, cycles
        code.append([line, int(code_entry.value), to_line, loc, int(code_entry.index),
                    int(code_entry.line_num), int(code_entry.addr), 0, 0])

    jumps = {}
    for k in range(info.jumps_len):
        jumps[info.jumps[k].key] = info.jumps[k].value

    return code, jumps


def getWaves_binary(name):
    filename = os.path.abspath(str(name))
    info = SO.AnalyseBinary(filename.encode("utf-8"))

    isValid = info.flags & 0x1
    if isValid == 0:
        print('Invalid trace ', name)
        return ([], [], [], [], None, [])
    flags = "navi" if (info.flags & 0x2) else "vega"
    kernel_addr = [info.kernel_id_addr[k].to_v2_pc() for k in range(info.num_kernel_ids)]
    events = [deepcopy(info.perfevents[k]) for k in range(info.num_events)]
    occupancy = [int(info.occupancy[k]) for k in range(int(info.num_occupancy))]

    assert(((info.flags >> 3) & 0x1FFF == ATT_VERSION)) # Check ATT parser version

    traces_python = {}
    for T in range(info.num_traces):
        if info.tracesizes[T] > 2:
            id = info.traceIDs[T]
            traces_python[id] = Trace(id, int(info.tracesizes[T]), info.tracedata[T])

    waves_python = []
    for k in range(info.num_waves):
        if info.wavedata[k].instructions_size > 2:
            waves_python.append(PythonWave(info.wavedata[k]))

    SO.FreeBinary(info.binaryID)

    return (traces_python, waves_python, events, occupancy, flags, kernel_addr, info.flags & 0x4)


def getWaves_stitch(traces, code, jumps, flags, latency_map, hitcount_map, bIsAuto, codeobjservice):
    for id in traces.keys():
        traces[id].instructions = stitch.stitch(traces[id].instructions, code, jumps, flags, bIsAuto, codeobjservice)
        if len(code) > hitcount_map.size:
            hitcount_map = np.pad(hitcount_map, [0,len(code)-hitcount_map.size])
            latency_map = np.pad(latency_map, [0,len(code)-latency_map.size])
        if traces[id].instructions is not None:
            for inst in traces[id].instructions[0]:
                hitcount_map[inst.asmline] += inst.num_waves
                latency_map[inst.asmline] += inst.cycles
    return hitcount_map, latency_map


def persist(trace_file, SIMD, traces):
    trace = Path(trace_file).name
    simds, waves = [], []
    begin_time, end_time, timeline, instructions, trace_ids = [], [], [], [], []
    mem_ins, issued_ins, valu_ins, valu_stalls = [], [], [], []
    vmem_ins, vmem_stalls, flat_ins, flat_stalls = [], [], [], []
    lds_ins, lds_stalls, salu_ins, salu_stalls = [], [], [], []
    smem_ins, smem_stalls, br_ins, br_taken_ins, br_stalls = [], [], [], [], []

    for wave in SIMD:
        try:
            if wave.instructions is None or traces[wave.traceid].instructions is None:
                continue
        except:
            continue
        simds.append(wave.simd)
        waves.append(wave.wave_id)
        begin_time.append(wave.begin_time)
        end_time.append(wave.end_time)
        trace_ids.append(wave.traceid)
        mem_ins.append(wave.num_mem_instrs)
        issued_ins.append(wave.num_issued_instrs)
        valu_ins.append(wave.num_valu_instrs)
        valu_stalls.append(wave.num_valu_stalls)
        vmem_ins.append(wave.num_vmem_instrs)
        vmem_stalls.append(wave.num_vmem_stalls)
        flat_ins.append(wave.num_flat_instrs)
        flat_stalls.append(wave.num_flat_stalls)
        lds_ins.append(wave.num_lds_instrs)
        lds_stalls.append(wave.num_lds_stalls)
        salu_ins.append(wave.num_salu_instrs)
        salu_stalls.append(wave.num_salu_stalls)
        smem_ins.append(wave.num_smem_instrs)
        smem_stalls.append(wave.num_smem_stalls)
        br_ins.append(wave.num_branch_instrs)
        br_taken_ins.append(wave.num_branch_taken_instrs)
        br_stalls.append(wave.num_branch_stalls)
        timeline.append(wave.timeline)

        cc = 0
        insts = []
        skips = traces[wave.traceid].instructions[-1]
        try:
            for v in traces[wave.traceid].instructions[0]:
                while cc in skips:
                    cc += 1
                t = wave.instructions[cc]
                insts.append((t[0], v.type, 0, t[1], v.asmline))
                cc += 1
        except:
            pass # Incomplete waves
        instructions.append((insts,) + traces[wave.traceid].instructions[1:-1])

    df = {
        "name": [trace for _ in range(len(begin_time))],
        "id": [i for i in range(len(begin_time))],
        "simd": simds,
        "wave_slot": waves,
        "begin_time": begin_time,
        "end_time": end_time,
        "mem_ins": mem_ins,
        "issued_ins": issued_ins,
        "valu_ins": valu_ins,
        "valu_stalls": valu_stalls,
        "vmem_ins": vmem_ins,
        "vmem_stalls": vmem_stalls,
        "flat_ins": flat_ins,
        "flat_stalls": flat_stalls,
        "lds_ins": lds_ins,
        "lds_stalls": lds_stalls,
        "salu_ins": salu_ins,
        "salu_stalls": salu_stalls,
        "smem_ins": smem_ins,
        "smem_stalls": smem_stalls,
        "br_ins": br_ins,
        "br_taken_ins": br_taken_ins,
        "br_stalls": br_stalls,
        "timeline": timeline,
        "instructions": instructions,
        "traceids": trace_ids,
    }
    return df


def mem_max(array):
    mem_dict = {}
    for SE in array:
        for wave in SE:
            for inst in wave:
                try:
                    mem_dict[inst[0]][0] = max(mem_dict[inst[0]][0], inst[1])
                except:
                    mem_dict[inst[0]] = inst[1:]
                assert mem_dict[inst[0]][1] == inst[2]

    return mem_dict


def lgk(count):
    return "lgkmcnt({0})".format(count)


def vmc(count):
    return "vmcnt({0})".format(count)


def both_cnt(count):
    return lgk(count) + " " + vmc(count)


def insert_waitcnt(flight_count, assembly_code):
    flight_count = mem_max(flight_count)
    for key in sorted(flight_count):
        line_n = key
        (
            issue_amount,
            waitcnt_amount,
        ) = flight_count[key]
        if "vmcnt" in assembly_code[line_n] and "lgkmcnt" in assembly_code[line_n]:
            counter_type = both_cnt
        elif "vmcnt" in assembly_code[line_n]:
            counter_type = vmc
        elif "lgkmcnt" in assembly_code[line_n]:
            counter_type = lgk
        else:
            print("Error: Line mismatch")
            exit(-1)

        for count in range(waitcnt_amount + 1, issue_amount):
            print("Inserted line: " + str(line_n))
            as_index = line_n - count / (issue_amount + 1)
            assembly_code[as_index] = "\ts_waitcnt {0}\t\t; Timing analysis.".format(
                counter_type(count)
            )
            as_index += 0.5 / (issue_amount + 1)
            assembly_code[as_index] = "\ts_nop 0\t\t\t\t\t\t; Counters: " + str(
                issue_amount
            )

    return assembly_code


def gen_timelines(DBFILES):
    TIMELINES = [np.zeros(int(1E6), dtype=np.float32) for k in range(5)]
    TIME_RESOLUTION = 16
    for df in DBFILES:
        for T in range(len(df["timeline"])):
            timeline = df["timeline"][T]
            time_acc = 0
            tuples3 = [(0, df["begin_time"][T])] + [(int(t[0]), int(t[1])) for t in timeline]

            for state in tuples3:
                t_end = (time_acc + state[1])//TIME_RESOLUTION
                if t_end > 1E8:
                    print("Warning: Time limit reached for ", state[0], state[1])
                    break
                elif t_end > TIMELINES[state[0]].size:
                    TIMELINES[state[0]] = np.hstack(
                        [TIMELINES[state[0]], np.zeros_like(TIMELINES[state[0]])]
                    )
                TIMELINES[state[0]][time_acc//TIME_RESOLUTION : t_end] += 1
                time_acc += state[1]
    return TIMELINES


if __name__ == "__main__":
    pathenv = os.getenv("OUTPUT_PATH")
    if pathenv is None:
        pathenv = "."
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "assembly_code", help="Path to the assembly code. Must be the first parameter."
    )
    parser.add_argument(
        "--trace_file", help="Filter for trace files", default=None, type=str
    )
    parser.add_argument(
        "--att_kernel", help="Kernel file", type=str, default=os.path.join(pathenv, "*_kernel.txt")
    )
    parser.add_argument("--ports", help="Server and websocket ports, default: 8000,18000")
    parser.add_argument(
        "--mode",
        help="""ATT analysis modes:\n
                        off: Only run ATT collection, disable analysis.\n
                        file: dump json files to disk.""",
        type=str,
        default="off",
    )
    args = parser.parse_args()
    args.mode = args.mode.lower().split(',')

    CSV_MODE = False
    FILE_MODE = False
    if 'csv' in args.mode:
        CSV_MODE = True
    if 'file' in args.mode:
        FILE_MODE = True

    if not CSV_MODE and not FILE_MODE:
        print("Skipping analysis.")
        quit()

    if os.getenv("COUNTERS_PATH"):
        with open(os.getenv("COUNTERS_PATH"), "r") as f:
            lines = [l.split("//")[0] for l in f.readlines()]

            EVENT_NAMES = []
            clean = lambda x: x.split("=")[1].split(" ")[0].split("\n")[0]
            for line in lines:
                if "PERFCOUNTER_ID=" in line:
                    EVENT_NAMES += ["id: " + clean(line)]
            for line in lines:
                if "PERFCOUNTER=" in line:
                    EVENT_NAMES += [clean(line).split("SQ_")[1].lower()]

    att_kernel_list = glob.glob(args.att_kernel)

    if len(att_kernel_list) == 0:
        print("Could not find att output kernel:", args.att_kernel)
        quit()

    for att_kernel in att_kernel_list:
        print('Parsing:', att_kernel)
        att_kernel_f = []
        with open(att_kernel, 'r') as f:
            for line in f:
                att_kernel_f.append(line.split('\n')[0])

        # returns element index of the file:// or memory:// filepath in the string
        get_path_loc = lambda x: np.sum([len(m) for m in x.split(' ')[:3]])+3
        # adds the OUTPUT_PATH env variable to a filepath if necessary
        add_pathnv = lambda x: x[:get_path_loc(x)] + os.path.join(pathenv, x[get_path_loc(x):])
        # returns the memory address in the string
        get_addr = lambda x: int(x.split(' ')[0][2:], 16)
        # Sets a preference for 'file' paths to be added before 'memory' paths. Sorts by addr.
        get_addr_preference = lambda x: [0 if 'file' in x[get_path_loc(x):] else 1<<60][0]

        # Get the GPU id in the string
        gpu_id = int(att_kernel_f[0].split(' ')[2].split('GPU[')[1].split(']')[0])
        # Eliminame first line in the att_kernel txt file and adds the OUTPUT_PATH as needed
        att_kernel_f = [add_pathnv(p) if '.out' == p[-4:] else p for p in att_kernel_f[1:]]
        # Sorts the list of codeobj by address, with 'file' given preference
        att_kernel_f = sorted(att_kernel_f, key=lambda x: get_addr(x)+get_addr_preference(x))
        assembly_code = deepcopy(args.assembly_code)

        # Assembly parsing
        bIsAuto = False
        if assembly_code.lower().strip() == 'auto':
            bIsAuto = True
            code = [['; Begin ATT ASM', 100, 0, '', 0, 0, 0, 0, 0]]
            jumps = []
        elif not Path(assembly_code).is_file():
            print("Invalid assembly_code('{0}')!".format(assembly_code))
            sys.exit(1)
        else:
            code, jumps = parse_binary(assembly_code, att_kernel)

        # Trace Parsing
        trace_instance_name = att_kernel.split("_kernel.txt")[0]
        if args.trace_file is None:
            filenames = glob.glob(trace_instance_name + "_*.att")
        else:
            filenames = glob.glob(args.trace_file)

        if len(filenames) == 0:
            print("Could not find trace files for", att_kernel)
            continue

        print('Att kernel:', att_kernel)

        DBFILES = []
        EVENTS = []
        OCCUPANCY = []
        GFXV = []
        analysed_filenames = []
        occupancy_filenames = []
        kernel_addr = {}

        latency_map = np.zeros((len(code)), dtype=np.int64)
        hitcount_map = np.zeros((len(code)), dtype=np.int32)

        gc.collect()
        if bIsAuto:
            codeservice = service.CodeobjService(gpu_id, att_kernel_f, SO.classify_asm_line)
        else:
            codeservice = None

        for name in filenames:
            traces, waves, perfevents, occupancy, gfxv, addrs, ftrace = getWaves_binary(name)
            if gfxv is None:
                continue
            if CSV_MODE == False and ftrace == 0:
                print('Generating occupancy information')

            for id, addr in enumerate(addrs):
                kernel_addr[id] = addr
            if len(occupancy) > 1:
                OCCUPANCY.append( occupancy )
                occupancy_filenames.append(name)

            if np.sum([0]+[len(s.instructions) for id, s in traces.items()]) == 0:
                print("No traces from", name)
                continue

            hitcount_map, latency_map = getWaves_stitch(traces, code, jumps, gfxv, latency_map, hitcount_map, bIsAuto, codeservice)

            analysed_filenames.append(name)
            EVENTS.append(perfevents)
            DBFILES.append( persist(name, waves, traces) )
            GFXV.append(gfxv)

        gc.collect()
        for k in range(len(code)):
            code[k][-2] = int(hitcount_map[k])
            code[k][-1] = int(latency_map[k])
            try:
                code[k][-3] = codeservice.ToRawPC(code[k][-3])
            except:
                pass
            if k > 0 and code[k-1][-3] == 0:
                code[k-1][-3] = code[k][-3]

        if CSV_MODE:
            from att_to_csv import dump_csv
            dump_csv(code, trace_instance_name, bIsAuto)

        if FILE_MODE:
            try:
                dispatchNames = {id: codeservice.getSymbolName(addr) for id, addr in kernel_addr.items()}
            except:
                dispatchNames = {id: "#addr"+str(addr) for id, addr in kernel_addr.items()}
            drawinfo = {
                "TIMELINES": gen_timelines(DBFILES),
                "EVENTS": EVENTS,
                "EVENT_NAMES": EVENT_NAMES,
                "OCCUPANCY": OCCUPANCY,
                "ShaderNames": occupancy_filenames,
                "DispatchNames": dispatchNames
            }
            view_trace(
                code,
                DBFILES,
                analysed_filenames,
                0,
                gfxv,
                drawinfo,
                trace_instance_name
            )



#!/usr/bin/env python3

# Copyright © Advanced Micro Devices, Inc., or its affiliates.
# SPDX-License-Identifier:  MIT

import sys

if sys.version_info[0] < 3:
    raise Exception("Must be using Python 3")

from collections import defaultdict
from copy import deepcopy

MAX_STITCHED_TOKENS = 200000000
MAX_FAILED_STITCHES = 256

SKIP = 0
SMEM = 1
SALU = 2
VMEM = 3
FLAT = 4
LDS = 5
VALU = 6
JUMP = 7
NEXT = 8
IMMED = 9
BRANCH = 10
GETPC = 11
SETPC = 12
SWAPPC = 13
LANEIO = 14
PCINFO = 15
WAVE_ENDED = 16
DONT_KNOW = 100

# Keeps track of register states for hipcc-generated assembly
class RegisterWatchList:
    def __init__(self, labels, code, jump_map, insts):
        self.registers = {"v" + str(k): [[] for m in range(64)] for k in range(64)}
        for k in range(128):
            self.registers["s" + str(k)] = []
        self.labels = labels

        self.code = code
        self.jump_map = jump_map
        self.insts = insts

    def jump(self, as_line):
        return self.jump_map[as_line[2]]

    def getcode(self, line):
        return self.code[line], 1

    def getincrement(self, line):
        return 1

    def try_translate(self, tok):
        if tok[0] in ["s"]:
            return self.registers[self.range(tok)[0]]
        elif "@" in tok:
            return self.labels[tok.split("@")[0]] + 1

    def range(self, r):
        reg = r.split(":")
        if len(reg) == 1:
            return reg
        else:
            r0 = reg[0].split("[")
            return [r0[0] + str(k) for k in range(int(r0[1]), int(reg[1][:-1]) + 1)]

    def tokenize(self, line):
        return [
            u for u in [t.split(",")[0].strip() for t in line.split(" ")] if len(u) > 0
        ]

    def getpc(self, line, next_line):
        try:
            dst = line.split(" ")[1].strip()
            label_dests = []
            try:
                label_dests = next_line[0].split(", ")
            except:
                pass
            try:
                label_dests.append(next_line[0].split(", ")[-1].split("@")[0])
            except:
                pass

            for label_dst in label_dests:
                try:
                    cur_label = self.labels[label_dst]
                    for reg in self.range(dst):
                        self.registers[reg] = deepcopy(cur_label)
                except:
                    pass
        except:
            pass

    def swappc(self, line, line_num, inst_num):
        try:
            tokens = self.tokenize(line)
            dst = tokens[1]
            src = tokens[2]

            popped = deepcopy(self.registers[self.range(src)[0]])
            self.registers[self.range(dst)[0]] = line_num + 1
            return popped
        except:
            return -1

    def setpc(self, line, inst_num):
        try:
            src = line.split(' ')[1].strip()
            return deepcopy(self.registers[self.range(src)[0]])
        except:
            return -1

    def scratch(self, line):
        try:
            tokens = self.tokenize(line)
            if "_load" in tokens[0]:
                dst = tokens[1]
                src = tokens[3] + tokens[4]
            else:
                src = tokens[2]
                dst = tokens[3] + tokens[4]
            self.registers[dst] = deepcopy(self.registers[src])
        except:
            pass

    def move(self, line):
        try:
            tokens = self.tokenize(line)
            if tokens[2][0] in ["s", "d"] and tokens[1][0] in ["s", "d"]:
                self.registers[self.range(tokens[1])[0]] = deepcopy(
                    self.registers[self.range(tokens[2])[0]]
                )
        except:
            pass

    def updatelane(self, line):
        tokens = self.tokenize(line)
        try:
            if "v_readlane" in tokens[0]:
                self.registers[tokens[1]] = deepcopy(self.registers[tokens[2]][int(tokens[3])])
            elif "v_writelane" in tokens[0]:
                self.registers[tokens[1]][int(tokens[3])] = deepcopy(self.registers[tokens[2]])
        except:
            pass

    # Matches tokens in reverse order
    def try_match_swapped(self, i, line, increment):
        try:
            return self.insts[i + 1].type == self.code[line][1] and self.insts[i].type == self.code[line + 1][1]
        except:
            return False

# Translates PC values to instructions, for auto captured ISA
class PCTranslator:
    def __init__(self, insts, code, raw_code, reverse_map, codeservice):
        self.codeservice = codeservice

        self.insts = insts
        self.addrmap = {c[-3] : (c, self.codeservice.GetInstruction(c[-3])[3]) for c in code if c[-3] > 0}

        self.code = code
        self.raw_code = raw_code
        self.reverse_map = reverse_map
        self.jump_map = {c[-3] : self.getjump_loc(c) for c in code if c[1] == BRANCH}

    def jump(self, as_line):
        return self.jump_map[as_line[-3]]

    def addsymbol(self, addr):
        if addr in self.addrmap:
            return

        symbol = self.codeservice.getSymbolName(addr)
        if symbol is None:
            symbol = "Unkown symbol at 0x" + hex(addr)

        last_line = self.raw_code[-1]
        newline = ['; ' + symbol, DONT_KNOW, last_line[2], '', last_line[4], last_line[5], 0, 0, 0]
        self.raw_code.append(newline)

    def getcode(self, addr):
        try:
            return self.addrmap[addr]
        except Exception as ex:
            new_inst = self.codeservice.GetInstruction(addr)
            if new_inst and new_inst[3]: # Check returned size > 0
                last_line = self.raw_code[-1]
                newline = [new_inst[1], new_inst[0], len(self.raw_code), new_inst[2], last_line[4]+1, last_line[5]+1, addr, 0, 0]
                if new_inst[0] == BRANCH:
                    self.jump_map[addr] = self.getjump_loc(newline)
                self.addrmap[addr] = (newline, new_inst[3])

                next = len(self.code)
                self.reverse_map[addr] = len(self.raw_code)
                self.raw_code.append(newline)
                self.code.append(newline)
                return newline, new_inst[3]
            else:
                raise ex

    def jump(self, asm_line):
        try:
            return self.jump_map[asm_line[-3]]
        except:
            loc = self.getjump_loc(asm_line)
            self.jump_map[asm_line[-3]] = loc
            return loc

    def getjump_loc(self, asm_line):
        try:
            dest = int(asm_line[0].split(' ')[-1])
            if dest >= 32768: dest -= 65536
            return asm_line[-3] + 4*dest+4
        except:
            return -1

    def getincrement(self, addr):
        return self.getcode(addr)[1]

    def try_translate(self, tok):
        pass
    def range(self, r):
        pass
    def tokenize(self, line):
        pass
    def getpc(self, line, next_line):
        pass
    def swappc(self, line, line_num, inst_index):
        try:
            inst_pos = inst_index+1
            while self.insts[inst_pos].type != PCINFO:
                inst_pos += 1
            return self.getcode(self.insts[inst_pos].cycles)[0][-3]
        except:
            print('SWAPPC warning: Could not find addr for', inst_index, line)
            return -1
    def setpc(self, line, inst_index):
        try:
            inst_pos = inst_index+1
            while self.insts[inst_pos].type != PCINFO:
                inst_pos += 1
            return self.getcode(self.insts[inst_pos].cycles)[0][-3]
        except:
            print('SETPC warning: Could not find addr for', inst_index, line)
            return -1
    def scratch(self, line):
        pass
    def move(self, line):
        pass
    def updatelane(self, line):
        pass

    # Matches tokens in reverse order
    def try_match_swapped(self, i, addr, increment):
        try:
            return  self.insts[i + 1].type == self.getcode(addr)[0][1] and \
                    self.insts[i].type == self.getcode(addr + increment)[0][1]
        except Exception as e:
            return False


def stitch(insts, raw_code, jumps, gfxv, bIsAuto, codeservice):
    bGFX9 = gfxv == 'vega'

    result, i, loopCount = [], 0, defaultdict(int)

    SMEM_INST = []  # scalar memory
    VLMEM_INST = []  # vector memory load
    VSMEM_INST = []  # vector memory store
    FLAT_INST = []
    NUM_SMEM = 0
    NUM_VLMEM = 0
    NUM_VSMEM = 0
    NUM_FLAT = 0
    skipped_immed = 0

    mem_unroll = []
    flight_count = []

    labels = {}
    jump_map = [0]

    # Clean the code and remove comments
    code = [raw_code[0]]
    for c in raw_code[1:]:
        c = list(c)
        c[0] = c[0].split(";")[0].split("//")[0].strip()
        jump_map.append(len(code))

        if c[1] != DONT_KNOW:
            code.append(c)
        elif ":" in c[0]:
            labels[c[0].split(":")[0]] = len(code)

    reverse_map = {}
    if bIsAuto:
        for k, v in enumerate(jump_map):
            try:
                reverse_map[code[v][-3]] = k
            except:
                pass
    else:
        for k, v in enumerate(jump_map):
            reverse_map[v] = k

    jumps = {jump_map[j] + 1: j for j in jumps}

    # Checks if we have guaranteed ordering in memory operations
    smem_ordering = 0
    vlmem_ordering = 0
    vsmem_ordering = 0

    num_failed_stitches = 0
    loops = 0
    maxline = 0

    pcskip = []
    if bIsAuto:
        try:
            firstinst = insts[0]

            if firstinst.type != PCINFO:
                print('Warning: Waves without PCINFO')
                return None
            elif firstinst.cycles == 0:
                print('Info: Some waves started before the trace')
                return None

            watchlist = PCTranslator(insts, code, raw_code, reverse_map, codeservice)
            watchlist.addsymbol(firstinst.cycles)
            line = firstinst.cycles
            lineincrement = watchlist.getincrement(line)
        except KeyError as e:
            print('Warning: Waves from addr', hex(e.args[0]), 'have no codeobj info.')
            for i in range(len(insts)):
                insts[i].asmline = 0
            return [i for k, i in enumerate(insts) if i.type != PCINFO], [], [], [], 1, 0, [k for k, i in enumerate(insts) if i.type == PCINFO]
        except Exception as e:
            print('Unknown error', e)
            return None
    else:
        line = 0
        lineincrement = 1
        watchlist = RegisterWatchList(labels=labels, code=code, jump_map=jump_map, insts=insts)

    N = len(insts)

    while i < N and line >= 0 and loops < MAX_STITCHED_TOKENS:
        if insts[i].type == PCINFO:
            pcskip.append(i)
            i += 1
            continue
        loops += 1

        inst = insts[i]
        try:
            as_line, lineincrement = watchlist.getcode(line)
        except:
            break

        matched = True
        next = line + lineincrement

        if not bIsAuto:
            if '_mov_' in as_line[0]:
                watchlist.move(as_line[0])
            elif 'scratch_' in as_line[0]:
                watchlist.scratch(as_line[0])

        if as_line[1] == DONT_KNOW or (as_line[1] == SKIP and not bGFX9):
            matched = False
        elif as_line[1] == GETPC:
            try:
                watchlist.getpc(as_line[0], watchlist.getcode(next)[0])
                matched = inst.type in [SALU, JUMP]
            except:
                matched = False
        elif as_line[1] == LANEIO:
            watchlist.updatelane(as_line[0])
            matched = inst.type == VALU
        elif as_line[1] == SETPC:
            next = watchlist.setpc(as_line[0], i)
            matched = inst.type in [SALU, JUMP]
            i += 1
            pcskip.append(i)
            while bIsAuto and next < 0 and i+1 < len(insts):
                i += 1
                if insts[i].type == PCINFO:
                    pcskip.append(i)
                    next = watchlist.setpc(as_line[0], i-1)
                else:
                    inst.cycles += insts[i].cycles
            if next < 0:
                print('Jump to unknown location in line', as_line[0])
                break
        elif as_line[1] == SWAPPC:
            matched = inst.type in [SALU, JUMP]
            next = watchlist.swappc(as_line[0], line, i)
            i += 1
            pcskip.append(i)
            while bIsAuto and next < 0 and i+1 < len(insts):
                i += 1
                if insts[i].type == PCINFO:
                    next = watchlist.swappc(as_line[0], line, i-1)
                    pcskip.append(i)
                else:
                    inst.cycles += insts[i].cycles
            if next < 0:
                print('Jump to unknown location in line', as_line[0])
                break
        elif inst.type == as_line[1]:
            if line in jumps:
                loopCount[jumps[line] - 1] += 1
            num_inflight = NUM_FLAT + NUM_SMEM + NUM_VLMEM + NUM_VSMEM

            if inst.type == SMEM or inst.type == LDS:
                smem_ordering = 1 if inst.type == SMEM else smem_ordering
                SMEM_INST.append([reverse_map[line], num_inflight])
                NUM_SMEM += 1
            elif inst.type == VMEM or (inst.type == FLAT and "global_" in as_line[0]):
                inc_ordering = False
                if "flat_" in as_line[0]:
                    inc_ordering = True

                if not bGFX9 and "store" in as_line[0]:
                    VSMEM_INST.append([reverse_map[line], num_inflight])
                    NUM_VSMEM += 1
                    if inc_ordering:
                        vsmem_ordering = 1
                else:
                    VLMEM_INST.append([reverse_map[line], num_inflight])
                    NUM_VLMEM += 1
                    if inc_ordering:
                        vlmem_ordering = 1
            elif inst.type == FLAT:
                smem_ordering = 1
                vlmem_ordering = 1
                vsmem_ordering = 1
                FLAT_INST.append([reverse_map[line], num_inflight])
                NUM_FLAT += 1
            elif inst.type == IMMED and "s_wait" in as_line[0] and not "s_wait_alu" in as_line[0]:
                if "lgkmcnt" in as_line[0] or "dscnt" in as_line[0] or "kmcnt" in as_line[0]:
                    try:
                        wait_N = int(as_line[0].split("lgkmcnt(")[1].split(")")[0])
                    except:
                        wait_N = 0
                    flight_count.append([as_line[5], num_inflight, wait_N])
                    if wait_N == 0:
                        smem_ordering = 0
                    if smem_ordering == 0:
                        offset = len(SMEM_INST) - wait_N
                        mem_unroll.append(
                            [reverse_map[line], SMEM_INST[:offset] + FLAT_INST]
                        )
                        SMEM_INST = SMEM_INST[offset:]
                        NUM_SMEM = len(SMEM_INST)
                        FLAT_INST = []
                        NUM_FLAT = 0
                    else:
                        NUM_SMEM = min(max(wait_N - NUM_FLAT, 0), NUM_SMEM)
                        NUM_FLAT = min(max(wait_N - NUM_SMEM, 0), NUM_FLAT)
                    num_inflight = NUM_FLAT + NUM_SMEM + NUM_VLMEM + NUM_VSMEM

                if "vmcnt" in as_line[0] or "loadcnt" in as_line[0]:
                    try:
                        wait_N = int(as_line[0].split("vmcnt(")[1].split(")")[0])
                    except:
                        wait_N = 0
                    flight_count.append([as_line[5], num_inflight, wait_N])
                    if wait_N == 0:
                        vlmem_ordering = 0
                    if vlmem_ordering == 0:
                        offset = len(VLMEM_INST) - wait_N
                        mem_unroll.append(
                            [reverse_map[line], VLMEM_INST[:offset] + FLAT_INST]
                        )
                        VLMEM_INST = VLMEM_INST[offset:]
                        NUM_VLMEM = len(VLMEM_INST)
                        FLAT_INST = []
                        NUM_FLAT = 0
                    else:
                        NUM_VLMEM = min(max(wait_N - NUM_FLAT, 0), NUM_VLMEM)
                        NUM_FLAT = min(max(wait_N - NUM_VLMEM, 0), NUM_FLAT)
                    num_inflight = NUM_FLAT + NUM_SMEM + NUM_VLMEM + NUM_VSMEM

                if "vscnt" in as_line[0] or (bGFX9 and "vmcnt" in as_line[0]) or "storecnt" in as_line[0]:
                    try:
                        wait_N = int(as_line[0].split('vscnt(')[1].split(')')[0])
                    except:
                        try:
                            wait_N = int(as_line[0].split('vmcnt(')[1].split(')')[0])
                        except:
                            wait_N = 0
                    flight_count.append([as_line[5], num_inflight, wait_N])
                    if wait_N == 0:
                        vsmem_ordering = 0
                    if vsmem_ordering == 0:
                        offset = len(VSMEM_INST) - wait_N
                        mem_unroll.append(
                            [reverse_map[line], VSMEM_INST[:offset] + FLAT_INST]
                        )
                        VSMEM_INST = VSMEM_INST[offset:]
                        NUM_VSMEM = len(VSMEM_INST)
                        FLAT_INST = []
                        NUM_FLAT = 0
                    else:
                        NUM_VSMEM = min(max(wait_N - NUM_FLAT, 0), NUM_VSMEM)
                        NUM_FLAT = min(max(wait_N - NUM_VSMEM, 0), NUM_FLAT)
                    num_inflight = NUM_FLAT + NUM_SMEM + NUM_VLMEM + NUM_VSMEM

        elif inst.type == JUMP and as_line[1] == BRANCH:
            next = watchlist.jump(as_line)
            if next is None or next == 0:
                print("Jump to unknown location!", as_line)
                break
        elif inst.type == NEXT and as_line[1] == BRANCH:
            pass
        else:
            matched = False
            if watchlist.try_match_swapped(i, line, lineincrement):
                temp = insts[i]
                insts[i] = insts[i + 1]
                insts[i + 1] = temp
                next = line
            else:
                hasWait = "s_wait" in as_line[0]
                if hasWait or "_load_" in as_line[0]:
                    if skipped_immed > 0 and hasWait:
                        matched = True
                        skipped_immed -= 1
                    elif 's_waitcnt' in as_line[0] and 'scratch_' not in as_line[0]:
                        print('WARNING: Parsing terminated at:', as_line)
                        break

        if matched or as_line[1] != DONT_KNOW:
            if matched:
                inst.asmline = reverse_map[line]
                result.append(inst)
                i += 1
                num_failed_stitches = 0
            elif not bGFX9 and inst.type == IMMED and line != next and as_line[1] != SKIP:
                skipped_immed += 1
                inst.asmline = reverse_map[line]
                result.append(inst)
                if 's_barrier' in as_line[0]:
                    next = line + lineincrement
                i += 1
            else:
                num_failed_stitches += 1

        maxline = max(reverse_map[line], maxline)
        line = next

    N = max(N, 1)
    if i != N and (insts[i].type == WAVE_ENDED or i == N-1):
        print('Warning - Wave ended.')
    elif i < N:
        print('Warning - Stitching rate: '+str(i * 100 / N)+'% matched', i, ' of ', N)
        try:
            print(line, code[line])
        except:
            pass
    else:
        while line < len(code):
            if "s_endpgm" in code[line]:
                mem_unroll.append(
                    [reverse_map[line], SMEM_INST + VLMEM_INST + VSMEM_INST + FLAT_INST]
                )
                break
            line += 1
        print('Success: Parsed', i, 'tokens')

    return result, loopCount, mem_unroll, flight_count, maxline, len(result), pcskip



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pytest
import json
import os

from rocprofiler_sdk.pytest_utils.dotdict import dotdict
from rocprofiler_sdk.pytest_utils import collapse_dict_list
from rocprofiler_sdk.pytest_utils.perfetto_reader import PerfettoReader
from rocprofiler_sdk.pytest_utils.otf2_reader import OTF2Reader


def pytest_addoption(parser):
    parser.addoption(
        "--json-input",
        action="store",
        help="Path to JSON file.",
    )
    parser.addoption(
        "--pftrace-input",
        action="store",
        help="Path to Perfetto trace file.",
    )
    parser.addoption(
        "--otf2-input",
        action="store",
        help="Path to OTF2 trace file.",
    )


@pytest.fixture
def json_data(request):
    filename = request.config.getoption("--json-input")
    with open(filename, "r") as inp:
        return dotdict(collapse_dict_list(json.load(inp)))


@pytest.fixture
def pftrace_data(request):
    filename = request.config.getoption("--pftrace-input")
    return PerfettoReader(filename).read()[0]


@pytest.fixture
def otf2_data(request):
    filename = request.config.getoption("--otf2-input")
    if not os.path.exists(filename):
        raise FileExistsError(f"{filename} does not exist")
    return OTF2Reader(filename).read()[0]



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import csv
import pandas as pd
import pytest
import json
import os
import io

from rocprofiler_sdk.pytest_utils.dotdict import dotdict
from rocprofiler_sdk.pytest_utils import collapse_dict_list
from rocprofiler_sdk.pytest_utils.perfetto_reader import PerfettoReader
from rocprofiler_sdk.pytest_utils.otf2_reader import OTF2Reader


def pytest_addoption(parser):
    parser.addoption(
        "--json-input",
        action="store",
        help="Path to JSON file.",
    )
    parser.addoption(
        "--pftrace-input",
        action="store",
        help="Path to Perfetto trace file.",
    )
    parser.addoption(
        "--otf2-input",
        action="store",
        help="Path to OTF2 trace file.",
    )
    parser.addoption(
        "--csv-input",
        action="store",
        nargs="+",
        help="Paths to CSV files.",
    )
    parser.addoption(
        "--summary-input",
        action="store",
        help="Path to summary markdown file.",
    )

    pd.set_option("display.width", 2000)
    # increase debug display of pandas dataframes
    for itr in ["rows", "columns", "colwidth"]:
        pd.set_option(f"display.max_{itr}", None)


@pytest.fixture
def json_data(request):
    filename = request.config.getoption("--json-input")
    with open(filename, "r") as inp:
        return dotdict(collapse_dict_list(json.load(inp)))


@pytest.fixture
def pftrace_data(request):
    filename = request.config.getoption("--pftrace-input")
    return PerfettoReader(filename).read()[0]


@pytest.fixture
def otf2_data(request):
    filename = request.config.getoption("--otf2-input")
    if not os.path.exists(filename):
        raise FileExistsError(f"{filename} does not exist")
    return OTF2Reader(filename).read()[0]


@pytest.fixture
def csv_data(request):
    filenames = request.config.getoption("--csv-input")
    return [
        (filename, list(csv.DictReader(open(filename, "r")))) for filename in filenames
    ]


@pytest.fixture
def summary_data(request):
    filename = request.config.getoption("--summary-input")
    if not os.path.exists(filename):
        raise FileExistsError(f"{filename} does not exist")

    domains = {}
    with open(filename, "r") as inp:
        lines = [itr.strip() for itr in inp.readlines()]
        lines = [itr for itr in lines if itr and not itr.startswith("|--")]

    def rework(x):
        tmp = [itr.strip(" ") for itr in x.split("|")]
        tmp = [itr.strip() for itr in tmp if len(itr.strip()) > 0]
        if tmp[0] == "NAME":
            tmp = [f'"{itr}"' for itr in tmp]
        else:
            tmp[0] = f'"{tmp[0]}"'
            tmp[1] = f'"{tmp[1]}"'
        return ",".join(tmp)

    def process_current_domain(_name, _list):
        if _name and _list:
            _list = [rework(itr) for itr in _list]
            _contents = "{}\n".format("\n".join(_list))
            ifs = io.StringIO(_contents)
            df = pd.read_csv(ifs)
            domains[_name] = df
            _name = None
            _list = []
        return (None, [])

    current_name = None
    current_list = []

    for itr in lines:
        if not itr.startswith("|") or itr.startswith("ROCPROFV3 "):
            current_name, current_list = process_current_domain(
                current_name, current_list
            )
            current_name = itr.strip().strip(":").replace("ROCPROFV3 ", "", 1)
            rpos = current_name.rfind(" SUMMARY")
            if rpos >= 0:
                current_name = current_name[:rpos]
        else:
            current_list += [itr]

    process_current_domain(current_name, current_list)

    return domains



#!/usr/bin/env python3

# Copyright © Advanced Micro Devices, Inc., or its affiliates.
# SPDX-License-Identifier:  MIT

import sys

if sys.version_info[0] < 3:
    raise Exception("Must be using Python 3")

import numpy as np
from io import BytesIO
import matplotlib.pyplot as plt
from copy import deepcopy
import json

COUNTERS_MAX_CAPTURES = 1 << 12


class Readable:
    def __init__(self, jsonstring):
        self.jsonstr = json.dumps(jsonstring)
        self.seek = 0

    def read(self, length=0):
        if length <= 0:
            return self.jsonstr
        else:
            if self.seek >= len(self):
                self.seek = 0
                return None
            response = self.jsonstr[self.seek : self.seek + length]
            self.seek += length
            return bytes(response, "utf-8")

    def __len__(self):
        return len(self.jsonstr)


class FileBytesIO:
    def __init__(self, iobytes):
        self.iobytes = deepcopy(iobytes)
        self.seek = 0

    def __len__(self):
        return self.iobytes.getbuffer().nbytes

    def read(self, length=0):
        if length <= 0:
            return bytes(self.iobytes.getbuffer())
        else:
            if self.seek >= self.iobytes.getbuffer().nbytes:
                self.seek = 0
                return None
            response = self.iobytes.getbuffer()[self.seek : self.seek + length]
            self.seek += length
            return bytes(response)


def get_delta_time(events):
    try:
        CUS = [[e.time for e in events if e.cu == k and e.bank == 0] for k in range(16)]
        CUS = [np.asarray(c).astype(np.int64) for c in CUS if len(c) > 2]
        return np.min([np.min(abs(c[1:] - c[:-1])) for c in CUS])
    except:
        return 1


def draw_wave_metrics(selections, normalize, TIMELINES, EVENTS, EVENT_NAMES):
    plt.figure(figsize=(15, 4))

    delta_step = 8
    quad_delta_time = max(
        delta_step, int(0.5 + np.min([get_delta_time(events) for events in EVENTS]))
    )
    maxtime = (
        np.max([np.max([e.time for e in events]) for events in EVENTS]) / quad_delta_time
        + 1
    )

    if maxtime * delta_step >= COUNTERS_MAX_CAPTURES:
        delta_step = 1
    while maxtime >= COUNTERS_MAX_CAPTURES:
        quad_delta_time *= 2
        maxtime /= 2

    maxtime = int(min(maxtime * delta_step, COUNTERS_MAX_CAPTURES))
    event_timeline = np.zeros((16, maxtime), dtype=np.int32)
    print("Delta:", quad_delta_time)
    print("Max_cycles:", maxtime * quad_delta_time * 4 // delta_step)

    cycles = 4 * quad_delta_time // delta_step * np.arange(maxtime)
    kernel = len(EVENTS) * quad_delta_time

    for events in EVENTS:
        for e in range(len(events) - 1):
            bk = events[e].bank * 4
            start = events[e].time // (quad_delta_time // delta_step)
            end = start + delta_step
            event_timeline[bk : bk + 4, start:end] += np.asarray(
                events[e].toTuple()[1:5]
            )[:, None]
        start = events[-1].time
        event_timeline[bk : bk + 4, start : start + delta_step] += np.asarray(
            events[-1].toTuple()[1:5]
        )[:, None]

    event_timeline = [
        np.convolve(e, [kernel for k in range(3)])[1:-1] for e in event_timeline
    ]
    # event_timeline = [e/kernel for e in event_timeline]

    if normalize:
        event_timeline = [100 * e / max(e.max(), 1e-5) for e in event_timeline]

    colors = [
        "blue",
        "green",
        "gray",
        "red",
        "orange",
        "cyan",
        "black",
        "darkviolet",
        "yellow",
        "darkred",
        "pink",
        "lime",
        "gold",
        "tan",
        "aqua",
        "olive",
    ]
    [
        plt.plot(cycles, e, "-", label=n, color=c)
        for e, n, c, sel in zip(event_timeline, EVENT_NAMES, colors, selections)
        if sel
    ]

    plt.legend()
    if normalize:
        plt.ylabel("As % of maximum")
    else:
        plt.ylabel("Value")
    plt.xlabel("Cycle")
    plt.subplots_adjust(left=0.04, right=1, top=1, bottom=0.1)

    figure_bytes = BytesIO()
    plt.savefig(figure_bytes, dpi=150)
    return EVENT_NAMES, FileBytesIO(figure_bytes)


def draw_wave_states(selections, normalize, TIMELINES):
    plot_indices = [1, 2, 3, 4]
    STATES = [["Empty", "Idle", "Exec", "Wait", "Stall"][k] for k in plot_indices]
    colors = [["gray", "orange", "green", "red", "blue"][k] for k in plot_indices]

    plt.figure(figsize=(15, 4))


    maxtime = max([np.max((TIMELINES[k]!=0)*np.arange(0,TIMELINES[k].size)) for k in plot_indices])
    maxtime = max(maxtime, 1)
    timelines = [deepcopy(TIMELINES[k][:maxtime]) for k in plot_indices]
    timelines = [np.pad(t, [0, maxtime - t.size]) for t in timelines]

    if normalize:
        timelines = np.array(timelines) / np.maximum(np.sum(timelines, 0) * 1e-2, 1e-7)

    trim = max(maxtime // 5000, 1)
    cycles = np.arange(0, timelines[0].size // trim, 1) * trim
    timelines = [
        time[: trim * (time.size // trim)].reshape((-1, trim)).mean(-1)
        if len(time) > 0
        else cycles * 0
        for time in timelines
    ]
    kernsize = 15
    kernel = np.asarray([
        np.exp(-abs(10 * k / kernsize)) for k in range(-kernsize // 2, kernsize // 2 + 1)
    ])
    kernel /= np.sum(kernel)

    timelines = [
        np.convolve(time, kernel)[kernsize // 2 : -kernsize // 2]
        for time in timelines if len(time) > 0
    ]
    maxtime *= 16
    cycles *= 16
    [
        plt.plot(cycles, t, label="State " + s, linewidth=1.1, color=c)
        for t, s, c, sel in zip(timelines, STATES, colors, selections)
        if sel
    ]

    plt.legend()
    if normalize:
        plt.ylabel("Waves state %")
    else:
        plt.ylabel("Waves state total")
    plt.xlabel("Cycle")
    plt.ylim(-1)
    plt.xlim(-maxtime // 200, maxtime + maxtime // 200 + 1)
    plt.subplots_adjust(left=0.04, right=1, top=1, bottom=0.1)
    figure_bytes = BytesIO()
    plt.savefig(figure_bytes, dpi=150)
    return STATES, FileBytesIO(figure_bytes)


def draw_occupancy_per_dispatch(selections, normalize, OCCUPANCY, dispatchnames):
    plt.figure(figsize=(15, 4))
    maxtime = 1
    delta = 1

    for k in range(len(OCCUPANCY)):
        if len(OCCUPANCY[k]) <= 1:
            continue
        for ev in OCCUPANCY[k]:
            maxtime = max(maxtime, ev[0])

    NUM_DOTS = 1600 # number of points taken for graphing
    delta = max(1, maxtime // NUM_DOTS) # Spacing between data points. Waves will be averaged over this interval.
    # Holds occupancy data
    chart = np.zeros((len(dispatchnames), maxtime // delta + 2), dtype=np.float32)

    for occ in OCCUPANCY:
        if len(occ) <= 1:
            continue
        # Number of waves multiplied by number of events
        small_chart = np.zeros_like(chart)
        # Holds number of events in that time period, for averaging.
        norm_fact = np.zeros_like(chart)
        norm_fact += 1E-5

        # Holds last known state per dispatch
        current_time = [0 for k in range(len(dispatchnames))]
        # Holds occupancy per Dispatch
        total_value = [0 for k in range(len(dispatchnames))]

        for time, en, kid in occ:
            b = current_time[kid]
            e = max(b + 1, time // delta)
            small_chart[kid][b:e] += total_value[kid]
            norm_fact[kid][b:e] += 1

            # Enable = 1 means a new wave started, enable = 0 means a wave has ended on that kernel ID.
            total_value[kid] += 2*en - 1
            current_time[kid] = time // delta
        for small, norm, time, value in zip(small_chart, norm_fact, current_time, total_value):
            small[time] += value
            norm[time] += value

        chart += small_chart/norm_fact # small_chart / norm_fact is the mean number of waves a tthat time point

    for (id, name), occ in zip(dispatchnames.items(), chart):
        plt.plot(np.arange(occ.size) * delta * 8, occ, label=str(id)+'#'+name, linewidth=1.1)

    plt.legend()
    if normalize:
        plt.ylabel("Occupancy %")
    else:
        plt.ylabel("Occupancy total")
    plt.xlabel("Cycle")
    plt.ylim(-1)
    plt.xlim(-maxtime // 200, maxtime + maxtime // 200 + delta + 1)
    plt.subplots_adjust(left=0.04, right=1, top=1, bottom=0.1)
    figure_bytes = BytesIO()
    plt.savefig(figure_bytes, dpi=150)
    return dispatchnames, FileBytesIO(figure_bytes)


def draw_occupancy(selections, normalize, OCCUPANCY, shadernames, numdispatchid):
    plt.figure(figsize=(15, 4))
    names = []

    g_maxtime = 1
    g_delta = 1
    for name, occ in zip(shadernames, OCCUPANCY):
        if len(occ) <= 1:
            continue

        occ_values = [0]
        occ_times = [0]

        for time, en, _ in occ:
            occ_times.append(time)
            occ_values.append(occ_values[-1] + 2*en - 1) # If enable = 1, increment. Else, decrement occupancy.

        try:
            names.append('SE'+name.split('_se')[1].split('.att')[0])
        except:
            names.append(name)

        NUM_DOTS = 1500 # Number of points taken for graphing
        maxtime = occ_times[-1]+1
        delta = max(1, maxtime // NUM_DOTS)
        g_maxtime = max(g_maxtime, maxtime)
        g_delta = max(g_delta, delta)
        chart = np.zeros((maxtime // delta + 1), dtype=np.float32)
        norm_fact = np.zeros_like(chart)
        norm_fact += 1E-6

        for i in range(len(occ_times)-1):
            b = occ_times[i] // delta
            e = max(b + 1, occ_times[i + 1] // delta)
            chart[b:e] += occ_values[i]
            norm_fact[b:e] += 1

        chart /= norm_fact
        if normalize:
            chart /= max(chart.max(), 1e-6)

        plt.plot(np.arange(chart.size) * delta, chart, label=names[-1], linewidth=1.1)

    plt.legend()
    if normalize:
        plt.ylabel("Occupancy %")
    else:
        plt.ylabel("Occupancy total")
    plt.xlabel("Cycle")
    plt.ylim(-1)
    plt.xlim(-g_maxtime // 200, g_maxtime + g_maxtime // 200 + g_delta + 1)
    plt.subplots_adjust(left=0.04, right=1, top=1, bottom=0.1)
    figure_bytes = BytesIO()
    plt.savefig(figure_bytes, dpi=150)
    return names, FileBytesIO(figure_bytes)


def getocc(u):
    # Parser struct occupancy_info_t
    # Bits 23:63 Time= Time divided by 8
    # Bit 18 = Enable (Wave start if 1, Wave end if 0)
    # Bits 0:11 is the kernel ID running on that wave
    return 8*int(u>>23), (u>>18) & 1, u&0xFFF


def GeneratePIC(drawinfo, selections=[True for k in range(16)], normalize=False):
    EVENTS = drawinfo["EVENTS"]

    response = {}
    figures = {}

    OCCUPANCY = drawinfo["OCCUPANCY"]
    # Transforms returned data into a array of events with each event being a tuple (time, enable, kernel ID)
    OCCUPANCY = [[getocc(u) for u in OCCUPANCY[k]] for k in range(len(OCCUPANCY))]

    states, figure = draw_occupancy(selections, normalize, OCCUPANCY, drawinfo["ShaderNames"], len(drawinfo["DispatchNames"]))
    response["occupancy.png"] = states
    figures["occupancy.png"] = figure

    states, figure = draw_occupancy_per_dispatch(selections, normalize, OCCUPANCY, drawinfo["DispatchNames"])
    response["dispatches.png"] = states
    figures["dispatches.png"] = figure

    states, figure = draw_wave_states(selections, normalize, drawinfo["TIMELINES"])
    response["timeline.png"] = states
    figures["timeline.png"] = figure

    if len(EVENTS) > 0 and np.sum([len(e) for e in EVENTS]) > 32:
        EVENT_NAMES, figure = draw_wave_metrics(
            selections, normalize, drawinfo["TIMELINES"], EVENTS, drawinfo["EVENT_NAMES"]
        )
        response["counters.png"] = EVENT_NAMES
        figures["counters.png"] = figure

    return Readable(response), figures



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest
import json

from collections import defaultdict


# helper function
def node_exists(name, data, min_len=1):
    assert name in data
    assert data[name] is not None
    if isinstance(data[name], (list, tuple, dict, set)):
        assert len(data[name]) >= min_len


def get_operation(record, kind_name, op_name=None):
    for idx, itr in enumerate(record["strings"]["buffer_records"]):
        if kind_name == itr["kind"]:
            if op_name is None:
                return idx, itr["operations"]
            else:
                for oidx, oname in enumerate(itr["operations"]):
                    if op_name == oname:
                        return oidx
    return None


def test_scratch_memory(json_input_data, csv_input_data):
    data = json_input_data["rocprofiler-sdk-tool"]
    buffer_records = data["buffer_records"]

    scratch_memory_data = buffer_records["scratch_memory"]

    _, bf_op_names = get_operation(data, "SCRATCH_MEMORY")

    assert len(bf_op_names) == 4

    scratch_reported_agent_ids = set()
    detected_agents_ids = set(
        agent["id"]["handle"] for agent in data["agents"] if agent["type"] == 2
    )

    assert len(scratch_memory_data) > 0

    # check buffering data
    for node in scratch_memory_data:
        assert "size" in node
        assert "kind" in node
        assert "flags" in node
        assert "thread_id" in node
        assert "end_timestamp" in node
        assert "start_timestamp" in node
        assert "allocation_size" in node

        assert "queue_id" in node
        assert "agent_id" in node
        assert "operation" in node
        assert "handle" in node["queue_id"]

        assert node.size > 0
        assert node.thread_id > 0
        assert node.agent_id.handle > 0
        assert node.queue_id.handle > 0
        assert node.start_timestamp > 0
        assert node.end_timestamp > 0
        assert node.start_timestamp < node.end_timestamp

        # validation for allocation size based on operation
        operation = bf_op_names[node["operation"]]
        if operation == "SCRATCH_MEMORY_FREE":
            # For free events, allocation size must be exactly 0
            assert (
                node["allocation_size"] == 0
            ), f"Free operation should have allocation_size=0, got {node['allocation_size']}"
        elif operation == "SCRATCH_MEMORY_ALLOC":
            # Fixme: For alloc events, must be > 0 and < 32GB
            assert (
                node["allocation_size"] > 0
            ), f"Alloc operation should have allocation_size > 0, got {node['allocation_size']}"
            assert (
                node["allocation_size"] < 32000000000
            ), f"Alloc operation size should be < 32GB, got {node['allocation_size']}"

        assert data.strings.buffer_records[node.kind].kind == "SCRATCH_MEMORY"
        assert (
            data.strings.buffer_records[node.kind].operations[node.operation]
            in bf_op_names
        )

        scratch_reported_agent_ids.add(node["agent_id"]["handle"])

    verify_scratch_memory_alternating_pattern(scratch_memory_data, bf_op_names)
    assert 2**64 - 1 not in scratch_reported_agent_ids
    assert scratch_reported_agent_ids == detected_agents_ids

    assert len(csv_input_data) > 0

    for row in csv_input_data:
        assert "Kind" in row, "Kind header not present in csv for scratch memory trace."
        assert (
            "Operation" in row
        ), "Operation header not present in csv for scratch memory trace."
        assert (
            "Agent_Id" in row
        ), "Agent_Id header not present in csv for scratch memory trace."
        assert (
            "Queue_Id" in row
        ), "Queue_Id header not present in csv for scratch memory trace."
        assert (
            "Thread_Id" in row
        ), "Thread_Id header not present in csv for scratch memory trace."
        assert (
            "Allocation_Size" in row
        ), "Allocation_Size header not present in csv for scratch memory trace."
        assert (
            "Alloc_Flags" in row
        ), "Alloc_Flags header not present in csv for scratch memory trace."
        assert (
            "Start_Timestamp" in row
        ), "Start_Timestamp header not present in csv for scratch memory trace."
        assert (
            "End_Timestamp" in row
        ), "End_Timestamp header not present in csv for scratch memory trace."

        assert row["Kind"] == "SCRATCH_MEMORY"
        assert row["Operation"] in ["SCRATCH_MEMORY_ALLOC", "SCRATCH_MEMORY_FREE"]
        assert int(row["Agent_Id"].split(" ")[-1]) >= 0
        assert int(row["Queue_Id"]) > 0
        assert int(row["Thread_Id"]) > 0
        assert int(row["Allocation_Size"]) >= 0
        assert int(row["Start_Timestamp"]) > 0
        assert int(row["End_Timestamp"]) > 0
        assert int(row["Start_Timestamp"]) < int(row["End_Timestamp"])


def verify_scratch_memory_alternating_pattern(scratch_memory_data, bf_op_names):
    """
    Verify that operations follow ALLOC→FREE→ALLOC→FREE pattern per (thread, flags) combination.
    """
    # Track operations by thread and flags
    thread_flag_operations = {}

    for node in scratch_memory_data:
        thread_id = node["thread_id"]
        operation = node["operation"]  # Numeric (1=ALLOC, 2=FREE)
        flags = node["flags"]
        timestamp = node["start_timestamp"]

        key = (thread_id, flags)
        if key not in thread_flag_operations:
            thread_flag_operations[key] = []

        thread_flag_operations[key].append((timestamp, operation))

    # Verify proper alternating sequence for each thread+flags combination
    for (thread_id, flags), operations in thread_flag_operations.items():
        # Sort by timestamp to ensure chronological order
        sorted_ops = [op for _, op in sorted(operations)]

        # Must start with ALLOC (operation code 1)
        if sorted_ops and sorted_ops[0] != 1:
            raise AssertionError(
                f"Thread {thread_id}, Flags {flags}: Must start with ALLOC, found operation code {sorted_ops[0]}"
            )

        # Check for alternating pattern - expected pattern is ALLOC→FREE→ALLOC→FREE
        for i in range(len(sorted_ops)):
            expected = 1 if i % 2 == 0 else 2  # 1=ALLOC, 2=FREE
            if sorted_ops[i] != expected:
                op_name = (
                    bf_op_names[sorted_ops[i]]
                    if sorted_ops[i] < len(bf_op_names)
                    else f"Unknown({sorted_ops[i]})"
                )
                expected_name = (
                    bf_op_names[expected]
                    if expected < len(bf_op_names)
                    else f"Unknown({expected})"
                )

                raise AssertionError(
                    f"Thread {thread_id}, Flags {flags}: Operation #{i+1} should be {expected_name} (code {expected}), found {op_name} (code {sorted_ops[i]})"
                )


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import absolute_import

import numpy as np
import pandas as pd
from collections import defaultdict
from .arbiter_state import validate_arbiter_state
from .other_instructions import (
    validate_valu_instructions,
    validate_flat_instructions,
    validate_lds_instructions,
)
from .s_instructions import (
    validate_internal_instructions,
    validate_barrier_instructions,
    validate_waitcnt,
    validate_branch_instructions,
    validate_scalar_instructions,
)

# Using Prefix Tree to classify the instruction type
# I did this instead of the regex becuase I wanted to try if we could
# generalize this approach for other types of instructions.
# The dream scenario: We have a giant list of all instructions and their
# types. Then we parse the list and dynamically determine the checks
# based on the instruction types.


# TODO: extract this outside of the file
class TrieNode:
    def __init__(self):
        self.children = {}
        self.instruction_type = None  # Store the instruction type at the leaf node


class PrefixTree:
    def __init__(self):
        self.root = TrieNode()

    def insert(self, full_prefix, instruction_type):
        """Insert a prefix and its associated instruction type into the Trie."""
        node = self.root
        for char in full_prefix:
            if char not in node.children:
                node.children[char] = TrieNode()
            node = node.children[char]
        node.instruction_type = (
            instruction_type  # Assign the instruction type at the leaf
        )

    def get_instruction_type(self, instruction):
        """Get the list of instruction types based on the longest matching prefix."""
        node = self.root
        matched_types = []  # List to store matched types

        # Traverse the instruction one character at a time
        for char in instruction:
            if char not in node.children:
                break  # Stop if no match is found

            node = node.children[char]

            # If we reach a node that has an instruction type, store it
            if node.instruction_type:
                matched_types.append(node.instruction_type)

        return matched_types


instructions_with_types = [
    ("s_", "SCALAR"),  # Scalar instructions (general category)
    ("s_waitcnt", "WAITCNT"),  # WAITCNT (specific)
    ("s_sendmsg", "MESSAGE"),  # MESSAGE (specific)
    ("s_barrier", "BARRIER"),  # BARRIER (specifix)
    ("s_swappc", "JUMP"),  # JUMP (specific)
    ("s_setpc", "JUMP"),  # JUMP
    ("s_setpc", "JUMP"),  # JUMP
    ("s_sleep", "JUMP"),  # JUMP
    ("s_branch", "BRANCH"),  # BRANCH
    ("s_cbranch", "BRANCH"),  # BRANCH (conditional)
    ("s_wakeup", "OTHER"),  # OHTER
    ("s_nop", "INTERNAL"),  # INTERNAL
    ("s_sleep", "INTERNAL"),  # INTERNAL
    ("v_", "VALU"),  # VALU
    ("v_mfma", "MATRIX"),  # MATRIX
    ("flat_", "FLAT"),  # FLAT
    ("global_", "FLAT"),  # FLAT
    ("ds_", "LDS"),  # LDS
    ("buffer_", "TEX"),  # TEX
]


inst_type_verify_functions = {
    "BRANCH": validate_branch_instructions,
    "WAITCNT": validate_waitcnt,
    # "OTHER": validate_other_instructions,
    "SCALAR": validate_scalar_instructions,
    "INTERNAL": validate_internal_instructions,
    # "JUMP": validate_jump_instructions,
    # "MESSAGE": validate_message_instructions,
    "BARRIER": validate_barrier_instructions,
    "VALU": validate_valu_instructions,
    "FLAT": validate_flat_instructions,
    "LDS": validate_lds_instructions,
}


def validate_stochastic_samples_json(data_json):
    # fill in the Prefix Tree
    prefix_tree = PrefixTree()
    for prefix, instruction_type in instructions_with_types:
        prefix_tree.insert(prefix, instruction_type)

    instructions = data_json["strings"]["pc_sample_instructions"]
    comments = data_json["strings"]["pc_sample_comments"]

    insts_per_prefix_type = defaultdict(list)

    for sample in data_json["buffer_records"]["pc_sample_stochastic"]:
        inst_index = sample["inst_index"]
        if inst_index == -1:
            # Ignoring samples from blit kernels
            continue
        record = sample["record"]
        # extend the record with the instruction
        record["inst"] = instructions[inst_index]

        # get the instruction type from prefix tree
        inst_prefix_types = prefix_tree.get_instruction_type(record["inst"])
        # each type must have a type
        assert len(inst_prefix_types) > 0
        # As more then one type can be matched, we take the last one as the most specific.
        inst_prefix_type = inst_prefix_types[-1]
        insts_per_prefix_type[inst_prefix_type].append(record)

        # For each sample, we need to validate wave_cnt and arbiter state
        wave_cnt = record["wave_cnt"]
        assert wave_cnt >= 0 and wave_cnt <= 32, "Invalid wave count"

        # arbiter state check
        snapshot = record["snapshot"]
        validate_arbiter_state(snapshot)

    # Check now the instruction type and arb state correlation.
    # We do that for all samples of a single instruction type all at once
    # to minimize the number of functions calls (one call for all samples, instead of a function
    # call per sample).
    # Please note that each sample is iterated at most twice.
    # The first time to group samples per instruction type, and the second time to validate samples.
    for inst_prefix_type, sample_records in insts_per_prefix_type.items():
        if inst_prefix_type in inst_type_verify_functions:
            inst_type_verify_functions[inst_prefix_type](sample_records)
        else:
            assert False, f"Unhandle instruction type: {inst_prefix_type}"



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2023-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import json
import os
import pytest
import pandas as pd

from rocprofiler_sdk.pytest_utils.dotdict import dotdict
from rocprofiler_sdk.pytest_utils import collapse_dict_list


def pytest_addoption(parser):
    parser.addoption(
        "--input-csv",
        action="store",
        help="Path to CSV file.",
    )

    parser.addoption(
        "--input-json",
        action="store",
        help="Path to CSV file.",
    )

    parser.addoption(
        "--all-sampled",
        action="store",
        help="All SW and HW units must be sampled.",
    )


@pytest.fixture
def input_csv(request):
    filename = request.config.getoption("--input-csv")
    if not os.path.isfile(filename):
        # The CSV file is not generated, because the dependency test
        # responsible to generate this file was skipped or failed.
        # Thus emit the message to skip this test as well.
        print("PC sampling unavailable")
    else:
        with open(filename, "r") as inp:
            return pd.read_csv(
                inp,
                na_filter=False,  # parse empty fields as ""
                keep_default_na=False,  # parse empty fields as ""
                dtype={
                    "Exec_Mask": "uint64",
                    "Instruction": str,
                    "Instruction_Comment": str,
                    "Wave_Issued_Instruction": bool,
                    "Instruction_Type": str,
                    "Stall_Reason": str,
                },
            )


@pytest.fixture
def input_json(request):
    filename = request.config.getoption("--input-json")
    with open(filename, "r") as inp:
        # Significant overhead of 5-6secs observed when feeding
        # data into the dotdict.
        # Using plain python dict instead
        return collapse_dict_list(json.load(inp))


@pytest.fixture
def all_sampled(request):
    _all_sampled_str = request.config.getoption("--all-sampled")
    return _all_sampled_str == "True"



#!/usr/bin/env python3
###############################################################################
# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

import os
import re

from .importer import RocpdImportData
from .query import export_sqlite_query
from .time_window import apply_time_window
from . import output_config
from . import libpyrocpd


def write_sql_query_to_csv(
    connection: RocpdImportData,
    config,
    query,
    filename="",
    postfix="trace",
) -> None:
    """Write the contents of a SQL query to a CSV file in the specified output path."""

    query_not_empty = f"""
        SELECT EXISTS (
            {query}
        )
    """

    # just return if the result is empty
    if not connection.execute(query_not_empty).fetchone()[0]:
        return

    # call query module to export to csv
    file_prefix = config.output_file + "_" if config.output_file else ""
    file_postfix = "_" + postfix if postfix else ""
    export_path = os.path.join(
        config.output_path, f"{file_prefix}{filename}{file_postfix}.csv"
    )

    kwargs = {"title_columns": True}
    export_sqlite_query(
        connection, query, export_format="csv", export_path=export_path, **kwargs
    )


def write_agent_info_csv(importData, config) -> None:

    # Define mapping of output column name to JSON key
    json_keys = [
        "node_id",
        "logical_node_id",
        "cpu_cores_count",
        "simd_count",
        "cpu_core_id_base",
        "simd_id_base",
        "max_waves_per_simd",
        "lds_size_in_kb",
        "gds_size_in_kb",
        "num_gws",
        "wave_front_size",
        "num_xcc",
        "cu_count",
        "array_count",
        "num_shader_banks",
        "simd_arrays_per_engine",
        "cu_per_simd_array",
        "simd_per_cu",
        "max_slots_scratch_cu",
        "gfx_target_version",
        "vendor_id",
        "device_id",
        "location_id",
        "domain",
        "drm_render_minor",
        "num_sdma_engines",
        "num_sdma_xgmi_engines",
        "num_sdma_queues_per_engine",
        "num_cp_queues",
        "max_engine_clk_ccompute",
        "max_engine_clk_fcompute",
        "sdma_fw_version.uCodeSDMA AS Sdma_Fw_Version",
        "fw_version.uCode AS Fw_Version",
        "cu_per_engine",
        "max_waves_per_cu",
        "workgroup_max_size",
        "family_id",
        "grid_max_size",
        "local_mem_size",
        "hive_id",
        "gpu_id",
        "workgroup_max_dim.x AS Workgroup_Max_Dim_X",
        "workgroup_max_dim.y AS Workgroup_Max_Dim_Y",
        "workgroup_max_dim.z AS Workgroup_Max_Dim_Z",
        "grid_max_dim.x AS Grid_Max_Dim_X",
        "grid_max_dim.y AS Grid_Max_Dim_Y",
        "grid_max_dim.z AS Grid_Max_Dim_Z",
        "vendor_name",
        "product_name",
    ]

    # Build SELECT clause for json_extract columns
    select_json = []
    for column in json_keys:
        match = re.match(r"(.+?)\s+AS\s+(.+)", column, re.IGNORECASE)
        column_name, column_alias = (
            (match.group(1), match.group(2)) if match else (column, column)
        )
        select_json.append(f"json_extract(extdata, '$.{column_name}') AS {column_alias}")

    capabilities = [
        "HotPluggable",
        "HSAMMUPresent",
        "SharedWithGraphics",
        "QueueSizePowerOfTwo",
        "QueueSize32bit",
        "QueueIdleEvent",
        "VALimit",
        "WatchPointsSupported",
        "WatchPointsTotalBits",
        "DoorbellType",
        "AQLQueueDoubleMap",
        "DebugTrapSupported",
        "WaveLaunchTrapOverrideSupported",
        "WaveLaunchModeSupported",
        "PreciseMemoryOperationsSupported",
        "DEPRECATED_SRAM_EDCSupport",
        "Mem_EDCSupport",
        "RASEventNotify",
        "ASICRevision",
        "SRAM_EDCSupport",
        "SVMAPISupported",
        "CoherentHostAccess",
        "DebugSupportedFirmware",
        "PreciseALUOperationsSupported",
        "PerQueueResetSupported",
    ]

    # Build SELECT clause for Capability columns
    select_capability = []
    for capability in capabilities:
        select_capability.append(
            f"json_extract(extdata, '$.capability.{capability}') AS Cap_{capability}"
        )

    # Add non-JSON columns
    fixed_keys = [
        "guid",
        "type AS Agent_Type",
        "name",
        "model_name",
    ]

    # to keep the right order
    select_clause = (
        fixed_keys[:1]
        + select_json[:2]
        + fixed_keys[1:2]
        + select_json[2:33]
        + select_capability
        + select_json[33:47]
        + fixed_keys[2:3]
        + select_json[47:]
        + fixed_keys[3:4]
    )

    select_clause = ",\n    ".join(select_clause)

    query = f"""
        SELECT
            {select_clause}
        FROM "rocpd_info_agent"
    """

    write_sql_query_to_csv(importData, config, query, "agent_info", "")


def build_agent_id_string(agent_index_value, prefix=""):

    agent_prefix = prefix + "_" if prefix else ""

    if agent_index_value == libpyrocpd.agent_indexing.node:  # absolute
        return f"'Agent ' || {agent_prefix}agent_abs_index"
    elif (
        agent_index_value == libpyrocpd.agent_indexing.logical_node
    ):  # relative (default)
        return f"'Agent ' || {agent_prefix}agent_log_index"
    elif (
        agent_index_value == libpyrocpd.agent_indexing.logical_node_type
    ):  # type-relative
        return f"{agent_prefix}agent_type || ' ' || {agent_prefix}agent_type_index"
    else:
        return ""


def write_kernel_csv(importData, config) -> None:

    agent_id = build_agent_id_string(config.agent_index_value)

    if config.kernel_rename:
        kernel_name = "region"
    else:
        kernel_name = "name"

    select_columns = [
        "guid",
        "'KERNEL_DISPATCH' AS Kind",
        f"{agent_id} AS Agent_Id",
        "queue_id",
        "stream_id",
        "tid AS Thread_Id",
        "dispatch_id",
        "kernel_Id",
        f"{kernel_name} AS Kernel_Name",
        "stack_id AS Correlation_Id",
        "start AS Start_Timestamp",
        "end AS End_Timestamp",
        "lds_size AS Lds_Block_Size",
        "scratch_size",
        "vgpr_count",
        "accum_vgpr_count",
        "sgpr_count",
        "workgroup_x AS Workgroup_Size_X",
        "workgroup_y AS Workgroup_Size_Y",
        "workgroup_z AS Workgroup_Size_Z",
        "grid_x AS Grid_Size_X",
        "grid_y AS Grid_Size_Y",
        "grid_z AS Grid_Size_Z",
    ]

    aliased_headers = []
    for column in select_columns:
        aliased_headers.append(column)

    select_clause = ",\n".join(aliased_headers)

    query = f"""
        SELECT
            {select_clause}
        FROM "kernels"
        ORDER BY
            guid ASC, start ASC, end DESC
    """
    write_sql_query_to_csv(importData, config, query, "kernel")


def write_memory_copy_csv(importData, config) -> None:

    src_agent_id = build_agent_id_string(config.agent_index_value, "src")
    dst_agent_id = build_agent_id_string(config.agent_index_value, "dst")

    query = f"""
        SELECT
            guid,
            'MEMORY_COPY' AS Kind,
            name AS Direction,
            stream_id,
            {src_agent_id} AS Source_Agent_Id,
            {dst_agent_id} AS Destination_Agent_Id,
            stack_id AS Correlation_Id,
            start AS Start_Timestamp,
            end AS End_Timestamp
        FROM "memory_copies"
        ORDER BY
            guid ASC, start ASC, end DESC
    """
    write_sql_query_to_csv(importData, config, query, "memory_copy")


def write_memory_allocation_csv(importData, config) -> None:

    agent_id = build_agent_id_string(config.agent_index_value)

    query = f"""
        SELECT
            guid,
            'MEMORY_ALLOCATION' AS Kind,
            CASE
                WHEN type = 'ALLOC'
                THEN 'MEMORY_ALLOCATION_ALLOCATE'
                ELSE 'MEMORY_ALLOCATION_' || type
            END AS Operation,
            CASE
                WHEN type != 'FREE'
                THEN {agent_id}
                ELSE '"'
            END AS Agent_Id,
            size AS Allocation_Size,
            '0x' || printf('%016X', address) AS Address,
            stack_id AS Correlation_Id,
            start AS Start_Timestamp,
            end AS End_Timestamp
        FROM "memory_allocations"
        ORDER BY
            guid ASC, start ASC, end DESC
    """
    write_sql_query_to_csv(importData, config, query, "memory_allocation")


def write_counters_csv(importData, config) -> None:

    agent_id = build_agent_id_string(config.agent_index_value)

    select_columns = [
        "guid",
        "stack_id AS Correlation_Id",
        "dispatch_id",
        f"{agent_id} AS Agent_Id",
        "queue_id",
        "pid AS Process_Id",
        "tid AS Thread_Id",
        "grid_size",
        "kernel_id",
        "kernel_name",
        "workgroup_size",
        "lds_block_size AS Lds_Block_Size",
        "scratch_size",
        "vgpr_count",
        "accum_vgpr_count",
        "sgpr_count",
        "counter_name",
        "value AS Counter_Value",
        "start AS Start_Timestamp",
        "end AS End_Timestamp",
    ]

    aliased_headers = []
    for column in select_columns:
        aliased_headers.append(column)

    select_clause = ",\n".join(aliased_headers)

    query = f"""
        SELECT
            {select_clause}
        FROM "counters_collection"
        ORDER BY
            guid ASC, start ASC, end DESC
    """
    write_sql_query_to_csv(importData, config, query, "counter_collection")


def write_scratch_memory_csv(importData, config) -> None:

    agent_id = build_agent_id_string(config.agent_index_value)

    query = f"""
        SELECT
            guid,
            'SCRATCH_MEMORY' AS Kind,
            'SCRATCH_MEMORY_' || operation AS Operation,
            {agent_id} AS Agent_Id,
            queue_id,
            tid AS Thread_Id,
            alloc_flags,
            start AS Start_Timestamp,
            end AS End_Timestamp
        FROM "scratch_memory"
        ORDER BY
            guid ASC, start ASC, end DESC
    """
    write_sql_query_to_csv(importData, config, query, "scratch_memory")


def write_region_csv(importData, config) -> None:

    query = """
        SELECT
            guid,
            category AS Domain,
            name AS Function,
            pid AS Process_Id,
            tid AS Thread_Id,
            stack_id AS Correlation_Id,
            start AS Start_Timestamp,
            end AS End_Timestamp
        FROM "regions"
        ORDER BY
            guid ASC, start ASC, end DESC
    """
    write_sql_query_to_csv(importData, config, query, "regions")


def write_csv(importData, config):

    write_agent_info_csv(importData, config)
    write_counters_csv(importData, config)
    write_kernel_csv(importData, config)
    write_memory_allocation_csv(importData, config)
    write_memory_copy_csv(importData, config)
    write_region_csv(importData, config)
    write_scratch_memory_csv(importData, config)


def execute(input, config=None, window_args=None, **kwargs):

    importData = RocpdImportData(input)

    apply_time_window(importData, **window_args)

    config = (
        output_config.output_config(**kwargs)
        if config is None
        else config.update(**kwargs)
    )

    write_csv(importData, config)


def add_args(parser):
    """Add csv arguments."""

    return []


def process_args(args, valid_args):
    ret = {}
    return ret


def main(argv=None):
    import argparse
    from .time_window import add_args as add_args_time_window
    from .time_window import process_args as process_args_time_window
    from .output_config import add_args as add_args_output_config
    from .output_config import process_args as process_args_output_config
    from .output_config import add_generic_args, process_generic_args

    parser = argparse.ArgumentParser(
        description="Convert rocPD to CSV files",
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    required_params = parser.add_argument_group("Required arguments")

    required_params.add_argument(
        "-i",
        "--input",
        required=True,
        type=output_config.check_file_exists,
        nargs="+",
        help="Input path and filename to one or more database(s), separated by spaces",
    )

    valid_out_config_args = add_args_output_config(parser)
    valid_generic_args = add_generic_args(parser)
    valid_time_window_args = add_args_time_window(parser)
    valid_csv_args = add_args(parser)

    args = parser.parse_args(argv)

    out_cfg_args = process_args_output_config(args, valid_out_config_args)
    generic_out_cfg_args = process_generic_args(args, valid_generic_args)
    window_args = process_args_time_window(args, valid_time_window_args)
    csv_args = process_args(args, valid_csv_args)

    all_args = {
        **out_cfg_args,
        **generic_out_cfg_args,
        **csv_args,
    }

    execute(args.input, window_args=window_args, **all_args)


if __name__ == "__main__":
    main()



# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import absolute_import


def validate_arbiter_state(snapshot):
    # VALU pipe checks
    if snapshot["dual_issue_valu"]:
        # (valu_issue = 1 & valu_stall = 0) is the only allowed
        assert (
            snapshot["arb_state_issue_valu"] == 1
            and snapshot["arb_state_stall_valu"] == 0
        ), "Dual issue VALU arbiter state check failed"
    else:
        # (valu_issue = 0 & value_stall = 1) is not allowed
        assert not (
            snapshot["arb_state_issue_valu"] == 0
            and snapshot["arb_state_stall_valu"] == 1
        ), "VALU arbiter state check failed"

    # Matrix pipe checks
    # matrix_issue = 0 & matrix_stall = 1 is not allowed
    assert not (
        snapshot["arb_state_issue_matrix"] == 0
        and snapshot["arb_state_stall_matrix"] == 1
    ), "Matrix arbiter state check failed"

    # scalar pipe checks
    # scalar_issue = 0 & scalar_stall = 1 is not allowed
    assert not (
        snapshot["arb_state_issue_scalar"] == 0
        and snapshot["arb_state_stall_scalar"] == 1
    ), "Scalar arbiter state check failed"

    # texture pipe checks
    # tex_issue = 0 & tex_stall = 1 is not allowed
    assert not (
        snapshot["arb_state_issue_vmem_tex"] == 0
        and snapshot["arb_state_stall_vmem_tex"] == 1
    ), "Texture arbiter state check failed"

    # LDS pipe checks
    # lds_issue = 0 & lds_stall = 1 is not allowed
    assert not (
        snapshot["arb_state_issue_lds"] == 0 and snapshot["arb_state_stall_lds"] == 1
    ), "LDS arbiter state check failed"

    # flat pipe checks
    # flat_issue = 0 & flat_stall = 1 is not allowed
    assert not (
        snapshot["arb_state_issue_flat"] == 0 and snapshot["arb_state_stall_flat"] == 1
    ), "Flat arbiter state check failed"

    # misc pipe checks
    # TODO: verify this
    # According to Joe's slides, the misc_stall cannot be 0.
    # However, the condition representing this case fails for `transpose` application
    #   assert((samples['Arbiter_State_Stall_Misc'] == 0).all())
    # Instead, I had to replace is with the condition belowe
    #   misc_issue = 0 & misc_stall = 1 is not allowed
    assert not (
        snapshot["arb_state_issue_misc"] == 0 and snapshot["arb_state_stall_misc"] == 1
    ), "Misc arbiter state check failed"

    # export pipe checks
    # We assume same conditions for Export pipe as for Misc (Joe's original),
    # so we should TODO: verify
    # exp_issue can take both 1 and 0, so no need to check it
    # exp_stall must be 0
    assert snapshot["arb_state_stall_exp"] == 0, "Export arbiter state check failed"

    # lds_direct pipe checks
    # This pipe doesn't exist on GFX9 so both issue and stall must be 0
    assert (
        snapshot["arb_state_issue_lds_direct"] == 0
    ), "LDS Direct arbiter state check failed"
    assert (
        snapshot["arb_state_stall_lds_direct"] == 0
    ), "LDS Direct arbiter state check failed"

    # brmsg pipe doesn't exist on GFX9 so both issue and stall must be 0
    assert snapshot["arb_state_issue_brmsg"] == 0, "BRMSG arbiter state check failed"
    assert snapshot["arb_state_stall_brmsg"] == 0, "BRMSG arbiter state check failed"



# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import absolute_import


def validate_valu_instructions(sample_records):
    allowed_stall_reasons = set(
        [
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL",
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_NO_INSTRUCTION_AVAILABLE",
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_NOT_WIN",
        ]
    )
    for record in sample_records:
        assert record["inst"].startswith("v_"), "VALU instruction must start with 'v_'"

        snapshot = record["snapshot"]
        if record["wave_issued"] == 1:
            # wave issued a VALU instruction
            assert record["inst_type"] == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_VALU"
            assert snapshot["arb_state_issue_valu"] == 1
            assert snapshot["arb_state_stall_valu"] == 0
        else:
            # wave did not issue a VALU instruction
            # inst_type is not relevant
            stall_reason = snapshot["stall_reason"]
            assert (
                stall_reason in allowed_stall_reasons
            ), "Invalid stall reason for VALU instruction"

            if (
                stall_reason
                == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL"
            ):
                assert snapshot["arb_state_issue_valu"] == 1
                # Expectation would be that the `arb_state_stall_valu` is 1, but in some examples,
                # I've observed different behavior.

            if (
                stall_reason
                == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_NOT_WIN"
            ):
                assert (
                    snapshot["arb_state_issue_valu"] == 1
                    or snapshot["arb_state_stall_matrix"] == 1
                ), "VALU or Matrix instruction should be issued"


def validate_flat_instructions(sample_records):
    allowed_stall_reasons = set(
        [
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL",
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_NO_INSTRUCTION_AVAILABLE",
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_NOT_WIN",
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ALU_DEPENDENCY",
        ]
    )
    for record in sample_records:
        assert record["inst"].startswith("flat_") or record["inst"].startswith(
            "global_"
        ), "Invalid name of FLAT instruction"

        snapshot = record["snapshot"]
        if record["wave_issued"] == 1:
            # wave issued a flat memory instruction
            assert (
                record["inst_type"] == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_FLAT"
            ), "Invalid instruction type for FLAT instruction"
            assert snapshot["arb_state_issue_flat"] == 1, "Arbiter issued flat"
            assert (
                snapshot["arb_state_stall_flat"] == 0
            ), "Arbiter should not stalled flat"

            # TODO: add checks when flat stalls LDS, and vice versa
            # If global_ inst, check ISSUE_FLAT=1, STALL_FLAT=0, ISSUE_LDS=1 -> STALL_LDS = 1
        else:
            # wave did not issue a flat instruction
            # inst_type is not relevant
            stall_reason = snapshot["stall_reason"]
            assert (
                stall_reason in allowed_stall_reasons
            ), "Invalid stall reason for flat instruction"

            if (
                stall_reason
                == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL"
            ):
                assert snapshot["arb_state_issue_flat"] == 1, "Arbiter issued flat"
                assert snapshot["arb_state_stall_flat"] == 1, "EX stalled flat"

            # In case of flat instructions, ARBITER_NOT_WIN might mean that
            # the FLAT/VMEM pipe was idle, so the flat instruction is issued to the arbiter
            # to wake up the clock in FLAT/VMEM, but cannot be issued to the execution pipeline.
            # Afterwards, the same instruction is reissued to the arbiter that sends it to the execution pipeline.
            # That's why `Arbiter_State_Issue_Flat` is not always true as in some other cases.


def validate_lds_instructions(sample_records):
    allowed_stall_reasons = set(
        [
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL",
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_NO_INSTRUCTION_AVAILABLE",
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_NOT_WIN",
            "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ALU_DEPENDENCY",
        ]
    )
    for record in sample_records:
        assert record["inst"].startswith("ds_"), "Invalid name of LDS instruction"

        snapshot = record["snapshot"]
        if record["wave_issued"] == 1:
            # wave issued an LDS memory instruction
            assert (
                record["inst_type"] == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_LDS"
            ), "Invalid instruction type for LDS instruction"
            assert snapshot["arb_state_issue_lds"] == 1, "Arbiter issued lds"
            assert snapshot["arb_state_stall_lds"] == 0, "EX should not stalled lds"

            # TODO: add checks when LDS stalls flat, and vice versa
            # ISSUE_LDS=1, STALL_LDS=0, ISSUE_FLAT=1 -> STALL_FLAT = 1
        else:
            # wave did not issue an LDS instruction
            # inst_type is not relevant
            stall_reason = snapshot["stall_reason"]
            assert (
                stall_reason in allowed_stall_reasons
            ), "Invalid stall reason for LDS instruction"

            if (
                stall_reason
                == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL"
            ):
                assert snapshot["arb_state_issue_lds"] == 1, "Arbiter issued flat"
                assert snapshot["arb_state_stall_lds"] == 1, "EX stalled flat"
            elif (
                stall_reason
                == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_NOT_WIN"
            ):
                assert snapshot["arb_state_issue_lds"] == 1, "Arbiter issued flat"



# MIT License
#
# Copyright (c) 2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import absolute_import

import numpy as np
import pandas as pd

from .s_instructions import validate_s_instructions
from .valu_instructions import validate_valu_instructions
from .texture_instructions import validate_texture_instructions
from .matrix_instructions import validate_matrix_instructions
from .lds_instructions import validate_lds_instructions
from .flat_instructions import validate_flat_instructions


def validate_wave_count(df):
    # Validating number of actives waves on a cu
    assert (
        (df["Wave_Count"] >= 1) & (df["Wave_Count"] <= 32)
    ).all(), "Invalid Wave_Count"


def validate_issued_instruction_type_no_inst(samples):
    # NO_INST type of instructions means instruction is not issued
    issued_type_no_inst = samples[samples["Instruction_Type"] == "NO_INST"]
    assert len(issued_type_no_inst) == 0, "NO_INST implies no instruction is issued"


def validate_issued_instruction_type_other(samples):
    # OTHER type of instructions still to be determined
    issued_type_other = samples[samples["Instruction_Type"] == "OTHER"]
    assert len(issued_type_other) == 0, "OTHER type of instruction observed first time"


def validate_issued_instruction_type_lds_direct(samples):
    # LDS_DIRECT type of instructions do not exist on gfx9
    issued_type_lds_direct = samples[samples["Instruction_Type"] == "LDS_DIRECT"]
    assert (
        len(issued_type_lds_direct) == 0
    ), "LDS direct type of instruction observed on GFX9"


def validate_issued_instruction_type_dual_valu(samples):
    # LDS_DIRECT type of instructions do not exist on gfx9
    issued_type_dual_valu = samples[samples["Instruction_Type"] == "DUAL_VALU"]
    assert (
        len(issued_type_dual_valu) == 0
    ), "DUAL_VALU type of instruction observed on GFX9"


# TODO: add checks for missing instruction types
# - export


def validate_stochastic_samples_csv(df: pd.DataFrame):
    # We expect mode valid than invalid samples
    # TODO: use stats for comparing valid vs invalid samples
    # invalid_samples = df[df["Valid"] == False]
    # valid_samples = df[df["Valid"]].copy()
    # assert len(valid_samples) > len(invalid_samples)

    # only valid samples reside in df
    valid_samples = df.copy()

    validate_wave_count(valid_samples)

    # The following checks assumes that we were able to decode
    # the instruction, meaning a code object and dispatch must be known.
    valid_samples = valid_samples[valid_samples["Dispatch_Id"] > 0]

    # scalar, barrier, waitcnt, jump, message, branches (taken and not taken)
    # are handled inside `validate_s_instructions` function
    validate_s_instructions(valid_samples)
    validate_valu_instructions(valid_samples)
    validate_texture_instructions(valid_samples)
    validate_matrix_instructions(valid_samples)
    validate_lds_instructions(valid_samples)
    validate_flat_instructions(valid_samples)

    # validating issued instructions for uncovered types
    valid_samples_issued = valid_samples[
        valid_samples["Wave_Issued_Instruction"] == True
    ].copy()
    validate_issued_instruction_type_no_inst(valid_samples_issued)
    validate_issued_instruction_type_other(valid_samples_issued)

    # The following two types of instructions should not be observed on gfx9
    validate_issued_instruction_type_lds_direct(valid_samples_issued)
    validate_issued_instruction_type_dual_valu(valid_samples_issued)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2023-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import sys
import pytest
import re
import os
import glob
import json


def test_json_data(json_data):
    data = json_data["rocprofiler-sdk-tool"]
    strings = data["strings"]
    assert "att_filenames" in strings.keys()
    att_files = data["strings"]["att_filenames"]
    assert len(att_files) > 0


def test_code_object_memory(code_object_file_path, json_data, output_path):

    data = json_data["rocprofiler-sdk-tool"]
    tool_memory_load = data["strings"]["code_object_snapshot_filenames"]
    gfx_pattern = "gfx[a-z0-9]+"
    match = re.search(gfx_pattern, tool_memory_load[1])
    assert match != None
    gpu_name = match.group(0)

    read_bytes = lambda filename: open(os.path.join(output_path, filename), "rb").read()
    # Loads all saved code objects
    tool_memory = [read_bytes(saved) for saved in tool_memory_load[1:]]

    found = False
    for hsa_file in code_object_file_path["hsa_memory_load"]:

        m = re.search(gfx_pattern, hsa_file)
        assert m != None
        gpu = m.group(0)

        if gpu == gpu_name:
            found = True
            hsa_memory_bytes = open(hsa_file, "rb").read()
            # Checks if hsa_file is one of the saved code objects
            assert any([hsa_memory_bytes == fs for fs in tool_memory])
            break
    assert found == True


def test_realtime_clock(output_path):

    def verify_sorted(timestamps):

        # Sort by shader_clock (index 0)
        timestamps_sorted = sorted(timestamps, key=lambda ts: ts[0])
        # Ensure realtime clock is non descreasing
        assert all(
            curr[1] >= prev[1]
            for prev, curr in zip(timestamps_sorted, timestamps_sorted[1:])
        )

    def verify_gfxclock(timestamps, rt_frequency):

        delta_shader_clock = timestamps[-1][0] - timestamps[0][0]
        delta_realtime_ts = timestamps[-1][1] - timestamps[0][1]
        gfxclock = rt_frequency * delta_shader_clock / delta_realtime_ts

        # gfxclock must be positive
        assert gfxclock > 0
        # gfxclock must be <10GHz
        assert gfxclock < 1e10

    pattern = os.path.join(output_path, "ui_output_*", "realtime.json")
    for rt_file in glob.glob(pattern):
        with open(rt_file, "r", encoding="utf-8") as f:
            json_file = json.load(f)

        frequency = json_file["metadata"]["frequency"]
        # frequency = 0 means aqlprofile is not instrumented
        if frequency > 0:
            for key, value in json_file.items():
                # Exclude metadata and single-clock timestamps
                if "metadata" not in key and len(value) >= 2:
                    verify_sorted(value)
                    verify_gfxclock(value, frequency)


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import absolute_import


def validate_matrix_instructions_issued(samples_issued):
    # issued instruction with type == MATRIX -> instruction starts with v_mfma
    issued_type_matrix = samples_issued[
        samples_issued["Instruction_Type"]
        == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_MATRIX"
    ]
    assert issued_type_matrix["Instruction"].apply(lambda x: x.startswith("v_mfma")).all()
    # v_mfma_f32 goes through Matrix (MAI) arbiter, while v_mfma_f64 goes through the VALU arbiter

    # SGEMM goes through Matrix (MAI arbiter)
    v_mfma_f32_issued = samples_issued[
        samples_issued["Instruction"].apply(lambda x: x.startswith("v_mfma_f32"))
    ]
    assert (
        v_mfma_f32_issued["Instruction_Type"]
        == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_MATRIX"
    ).all()

    # DGEMM goes through VALU arbiter
    v_mfma_f64_issued = samples_issued[
        samples_issued["Instruction"].apply(lambda x: x.startswith("v_mfma_f64"))
    ]
    assert (v_mfma_f64_issued["Instruction_Type"] == "MATRIX").all()
    assert len(issued_type_matrix) == len(v_mfma_f32_issued) + len(v_mfma_f64_issued)

    # TODO: find an example with MAI instructions


def validate_dgemm_matrix_instructions_stalled(samples):
    v_mfma_f64_samples = samples[
        samples["Instruction"].apply(lambda x: x.startswith("v_mfma_f64"))
    ]
    v_mfma_f64_stalled = v_mfma_f64_samples[
        v_mfma_f64_samples["Wave_Issued_Instruction"] == False
    ]

    assert (
        v_mfma_f64_stalled["Stall_Reason"]
        .apply(
            lambda x: x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL"
            or x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_NO_INSTRUCTION_AVAILABLE"
            or x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_NOT_WIN"
        )
        .all()
    )


def validate_sgemm_matrix_instructions_stalled(samples):
    v_mfma_f32_samples = samples[
        samples["Instruction"].apply(lambda x: x.startswith("v_mfma_f32"))
    ]
    v_mfma_f32_stalled = v_mfma_f32_samples[
        v_mfma_f32_samples["Wave_Issued_Instruction"] == False
    ]
    assert (
        v_mfma_f32_stalled["Stall_Reason"]
        .apply(
            lambda x: x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL"
            or x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_NO_INSTRUCTION_AVAILABLE"
            or x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_NOT_WIN"
        )
        .all()
    )


def validate_matrix_instructions_stalled(samples):
    validate_dgemm_matrix_instructions_stalled(samples)
    validate_sgemm_matrix_instructions_stalled(samples)
    # TODO" find an example to test this


def validate_matrix_instructions(samples):
    samples_issued = samples[samples["Wave_Issued_Instruction"]]
    validate_matrix_instructions_issued(samples_issued)
    validate_matrix_instructions_stalled(samples)



###############################################################################
# MIT License
#
# Copyright (c) 2023 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

import sys
import os

try:
    import ctypes

    sqlite3lib = ctypes.CDLL("libsqlite3.so")
except Exception:
    pass

try:
    import sqlite3

    sqlite3.connect(":memory:")  # Test if sqlite3 is available
except Exception:
    pass

from . import libpyrocpd
from .importer import RocpdImportData

__all__ = [
    "connect",
    "execute",
    "read_agents",
    "read_nodes",
    "read_processes",
    "read_threads",
    "write_perfetto",
    "write_csv",
    "write_otf2",
    "RocpdImportData",
    "version_info",
]

version_info = {
    "version": "1.1.0",
    "major": int("1"),
    "minor": int("1"),
    "patch": int("0"),
    "git_revision": "e1a6bc5663304b9c586b3254b8920f2981057804",
    "library_arch": "",
    "system_name": "Linux",
    "system_processor": "x86_64",
    "system_version": "6.19.8-arch1-1",
    "compiler_id": "GNU",
    "compiler_version": "15.2.1",
    "rocm_version": "7.2.1",
}


def format_path(path, tag=os.path.basename(sys.executable)):
    return libpyrocpd.format_path(path, tag)


def connect(input, *args, **kwargs):
    return RocpdImportData(input, *args, **kwargs)


def execute(data, *args, **kwargs):
    return data.execute(*args, **kwargs)


def read_agents(data, condition=""):
    return libpyrocpd.read_agents(data, condition)


def read_nodes(data, condition=""):
    return libpyrocpd.read_nodes(data, condition)


def read_processes(data, condition=""):
    return libpyrocpd.read_processes(data, condition)


def read_threads(data, condition=""):
    return libpyrocpd.read_threads(data, condition)


def write_perfetto(connection, config=None, **kwargs):
    """
    Write Perfetto pftrace output file

    Args:
        connection (rocpd.RocpdImportData):
            rocPD instance of database connection(s)
        config (rocpd.output_config.output_config):
            Output specification

    Returns:
        bool: returns True if successful
    """
    from . import output_config

    config = (
        output_config.output_config(**kwargs)
        if config is None
        else config.update(**kwargs)
    )

    return libpyrocpd.write_perfetto(connection, config)


def write_csv(connection, config=None, **kwargs):
    """
    Write CSV output file(s)

    Args:
        connection (rocpd.RocpdImportData):
            rocPD instance of database connection(s)
        config (rocpd.output_config.output_config):
            Output specification

    Returns:
        bool: returns True if successful
    """
    from . import output_config

    config = (
        output_config.output_config(**kwargs)
        if config is None
        else config.update(**kwargs)
    )

    return libpyrocpd.write_csv(connection, config)


def write_otf2(connection, config=None, **kwargs):
    """
    Write OTF@ output file

    Args:
        connection (rocpd.RocpdImportData):
            rocPD instance of database connection(s)
        config (rocpd.output_config.output_config):
            Output specification

    Returns:
        bool: returns True if successful
    """
    from . import output_config

    config = (
        output_config.output_config(**kwargs)
        if config is None
        else config.update(**kwargs)
    )

    return libpyrocpd.write_otf2(connection, config)



# MIT License
#
# Copyright (c) 2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import absolute_import


def validate_issued_instruction_type_branch_taken(samples):
    # issued instruction with type BRANCH_TAKEN -> instruction starts with either s_cbranch or s_branch
    issued_type_branch_taken = samples[
        (
            samples["Instruction_Type"]
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_BRANCH_TAKEN"
        )
        & samples["Wave_Issued_Instruction"]
    ]
    assert (
        issued_type_branch_taken["Instruction"]
        .apply(lambda x: x.startswith("s_branch") or x.startswith("s_cbranch"))
        .all()
    )
    assert issued_type_branch_taken["Wave_Issued_Instruction"].all()

    # if issued instruction starts with s_branch (unconditional branch) -> its type must be BRANCH_TAKEN
    issued_s_branch = samples[
        samples["Instruction"].apply(lambda x: x.startswith("s_branch"))
        & samples["Wave_Issued_Instruction"]
    ]
    assert (
        issued_s_branch["Instruction_Type"]
        == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_BRANCH_TAKEN"
    ).all()

    # see `validate_issued_instruction_type_branch_not_taken` for more info about s_cbranch checks


def validate_issued_instruction_type_branch_not_taken(samples):
    # issued instruction with type BRANCH_NOT_TAKEN -> instruction is conditional branch (starts s_cbranch)
    issued_type_branch_not_taken = samples[
        (
            samples["Instruction_Type"]
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_BRANCH_NOT_TAKEN"
        )
        & samples["Wave_Issued_Instruction"]
    ]
    assert (
        issued_type_branch_not_taken["Instruction"]
        .apply(lambda x: x.startswith("s_cbranch"))
        .all()
    )
    assert issued_type_branch_not_taken["Wave_Issued_Instruction"].all()

    # if issued instruction starts with s_cbranch -> its type is either BRANCH_TAKEN on BRANCH_NOT_TAKEN
    issued_s_cbranch = samples[
        samples["Instruction"].apply(lambda x: x.startswith("s_cbranch"))
        & samples["Wave_Issued_Instruction"]
    ]
    assert (
        (
            issued_s_cbranch["Instruction_Type"]
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_BRANCH_TAKEN"
        )
        | (
            issued_s_cbranch["Instruction_Type"]
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_TYPE_BRANCH_NOT_TAKEN"
        )
    ).all()


def s_branch_not_issued(stalled_samples):
    s_branch_stalled = stalled_samples[
        stalled_samples["Instruction"].apply(lambda x: x.startswith("s_branch"))
    ]

    if len(s_branch_stalled) > 0:
        # No ALUDEP nor ARBWINEXSTALL observed so far for unconditional branches
        assert (
            s_branch_stalled["Stall_Reason"]
            .apply(
                lambda x: x
                != "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ALU_DEPENDENCY"
                and x
                != "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL"
            )
            .all()
        )


def validate_stalled_branches(samples):
    stalled_samples = samples[samples["Wave_Issued_Instruction"] == False]

    assert (
        stalled_samples["Stall_Reason"]
        .apply(
            lambda x: x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ALU_DEPENDENCY"
            or x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_NO_INSTRUCTION_AVAILABLE"
            or x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_NOT_WIN"
            or x
            == "ROCPROFILER_PC_SAMPLING_INSTRUCTION_NOT_ISSUED_REASON_ARBITER_WIN_EX_STALL"
        )
        .all()
    )

    # Further constraints for unconditional branches
    s_branch_not_issued(stalled_samples)


def validate_branch_instructions(all_samples, branch_samples):
    """
    Use all_samples to verify the ROCProfV3 determines `Instruction_Type` field properly.

    Use filtered_samples to verify both issued and stalled branch instructions.
    """
    # For the issued branches, use all samples, as the called functions will do
    # separation based on branch type (conditional or unconditional)
    validate_issued_instruction_type_branch_taken(all_samples)
    validate_issued_instruction_type_branch_not_taken(all_samples)

    # stalled branches
    validate_stalled_branches(branch_samples)



#!/usr/bin/env python3
import sys

if sys.version_info[0] < 3:
    raise Exception("Must be using Python 3")

import http.server
import socketserver
import socket
import os
import sys


class NoCacheHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    def end_headers(self):
        self.send_my_headers()
        http.server.SimpleHTTPRequestHandler.end_headers(self)

    def send_my_headers(self):
        self.send_header("Cache-Control", "no-cache, no-store, must-revalidate")
        self.send_header("Pragma", "no-cache")
        self.send_header("Expires", "0")

    def do_GET(self):
        if ".png?" in self.path:
            self.path = self.path.split(".png?")[0] + ".png"

        http.server.SimpleHTTPRequestHandler.do_GET(self)


class RocTCPServer(socketserver.TCPServer):
    def server_bind(self):
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.server_address)


def run_server():
    Handler = NoCacheHTTPRequestHandler
    os.chdir(os.path.join(os.path.dirname(os.path.abspath(__file__)), "."))
    try:
        with RocTCPServer((IPAddr, PORT), Handler) as httpd:
            httpd.serve_forever()
    except KeyboardInterrupt:
        pass


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        hostname = socket.gethostname()
        IPAddr = socket.gethostbyname(hostname)
        s.connect(({IPAddr}, 1))
    except Exception:
        IPAddr = "127.0.0.1"
    finally:
        return IPAddr


IPAddr = get_ip()
PORT = 8000

if len(sys.argv) > 1:
    PORT = int(sys.argv[1])
print("serving at port: {0}".format(PORT))

try:
    run_server()
except KeyboardInterrupt:
    print("Exitting.")



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json
import pytest
import pandas as pd
import re

from rocprofiler_sdk.pytest_utils.dotdict import dotdict
from rocprofiler_sdk.pytest_utils import collapse_dict_list


def pytest_addoption(parser):

    parser.addoption(
        "--input-json-pass1",
        action="store",
        help="Path to JSON file.",
    )

    parser.addoption(
        "--input-json-pass2",
        action="store",
        help="Path to JSON file.",
    )

    parser.addoption(
        "--input-json-pass3",
        action="store",
        help="Path to JSON file.",
    )

    parser.addoption(
        "--json-config",
        action="store",
        help="Path to input JSON file.",
    )


def tokenize(kernel_iteration_range):
    range_str = kernel_iteration_range.replace("[", "").replace("]", "")
    split_list = range_str.split(",")
    _range = []
    for split_string in split_list:
        if "-" in split_string:
            interval = split_string.split("-")
            [
                _range.append(i)
                for i in list(range((int)(interval[0]), (int)(interval[1]) + 1))
            ]
        else:
            _range.append(int(split_string))
    return _range


def extract_iteration_list(jobs, pass_):

    kernel_iteration_range = jobs[pass_]["kernel_iteration_range"].strip()
    return tokenize(kernel_iteration_range)


def process_config(out_file, input_config, pass_):

    ret_dict = {}

    with open(out_file, "r") as inp:
        ret_dict["json_data"] = dotdict(collapse_dict_list(json.load(inp)))

    with open(input_config, "r") as inp:
        jobs = dotdict(collapse_dict_list(json.load(inp)))["jobs"]
        ret_dict["iteration_range"] = extract_iteration_list(jobs, pass_)

    return ret_dict


@pytest.fixture
def input_json_pass1(request):
    out_file = request.config.getoption("--input-json-pass1")
    input_config = request.config.getoption("--json-config")
    return process_config(out_file, input_config, 0)


@pytest.fixture
def input_json_pass2(request):
    out_file = request.config.getoption("--input-json-pass2")
    input_config = request.config.getoption("--json-config")
    return process_config(out_file, input_config, 1)


@pytest.fixture
def input_json_pass3(request):
    out_file = request.config.getoption("--input-json-pass3")
    input_config = request.config.getoption("--json-config")
    return process_config(out_file, input_config, 2)
